"""Defines 2D and 3D data-structures storing vectors of integers.

This module defines several classes to store pairs (2D) or 3-tuples (3D) of
integers representing coordinates. They are used all over the code base to avoid
any coordinate system mess and to explicitly annotate each coordinate with its
significance. Basically, instead of having

.. code-block:: python

    coords = (0, 4)  # Is it (y, x) or (x, y)?
    # 0 used below, but if the coords tuple was obtained from a call to
    # `.shape` from a numpy array it should rather be a 1.
    x = coords[0]

we have

.. code-block:: python

    coords = Position2D(0, 4)
    x = coords.x

This is particularly useful when we, as humans, are mostly used to always have
(x, y) coordinates but some libraries (such as numpy) reverse that order for
indexing.

"""

from __future__ import annotations

import re
from dataclasses import astuple, dataclass
from enum import Enum

import numpy as np
import numpy.typing as npt
from typing_extensions import Self

from tqec.utils.exceptions import TQECError


@dataclass(frozen=True, order=True)
class Vec2D:
    x: int
    y: int


@dataclass(frozen=True, order=True)
class Vec3D:
    x: int
    y: int
    z: int


class Position2D(Vec2D):
    """Represent a position on a 2-dimensional plane.

    Warning:
        This class represents a position without any knowledge of the coordinate
        system being used. As such, it should only be used when the coordinate
        system is meaningless or in localised places where the coordinate system
        is obvious. In particular, this class should be avoided in interfaces.

    """

    def with_block_coordinate_system(self) -> BlockPosition2D:
        """Return a :class:`.BlockPosition2D` from ``self``."""
        return BlockPosition2D(self.x, self.y)  # pragma: no cover

    def is_neighbour(self, other: Position2D) -> bool:
        """Check if the other position is near to this position, i.e. Manhattan distance is 1."""
        return abs(self.x - other.x) + abs(self.y - other.y) == 1

    def to_3d(self, z: int = 0) -> Position3D:
        """Get a 3-dimensional position with the ``x`` and ``y`` coordinates from ``self``."""
        return Position3D(self.x, self.y, z)


class PhysicalQubitPosition2D(Position2D):
    """Represent the position of a physical qubit on a 2-dimensional plane."""


class PlaquettePosition2D(Position2D):
    """Represent the position of a plaquette on a 2-dimensional plane."""

    def get_origin_position(self, shift: Shift2D) -> PhysicalQubitPosition2D:
        """Return the position of the plaquette origin."""
        return PhysicalQubitPosition2D(shift.x * self.x, shift.y * self.y)  # pragma: no cover


class BlockPosition2D(Position2D):
    """Represent the position of a block on a 2-dimensional plane."""

    def get_top_left_plaquette_position(self, block_shape: PlaquetteShape2D) -> PlaquettePosition2D:
        """Return the position of the top-left plaquette of the block."""
        return PlaquettePosition2D(block_shape.x * self.x, block_shape.y * self.y)


class Shape2D(Vec2D):
    def to_numpy_shape(self) -> tuple[int, int]:
        """Return the shape according to numpy indexing.

        In the coordinate system used in this library, numpy indexes arrays using (y, x)
        coordinates. This method is here to translate a Shape instance to a numpy shape
        transparently for the user.

        In the coordinate system used in this library, numpy indexes arrays using (y, x)
        coordinates. This method is here to translate a ``Shape`` instance to a numpy shape
        transparently for the user.

        """
        return (self.y, self.x)


class PlaquetteShape2D(Shape2D):
    """Represent a 2-dimensional shape using plaquette coordinate system."""


class PhysicalQubitShape2D(Shape2D):
    """Represent a 2-dimensional shape using physical qubit coordinate system."""


class Shift2D(Vec2D):
    def __mul__(self, factor: int) -> Shift2D:
        return Shift2D(factor * self.x, factor * self.y)  # pragma: no cover

    def __rmul__(self, factor: int) -> Shift2D:
        return self.__mul__(factor)  # pragma: no cover


class Position3D(Vec3D):
    """A 3D integer position."""

    x: int
    y: int
    z: int

    def shift_by(self, dx: int = 0, dy: int = 0, dz: int = 0) -> Self:
        """Shift the position by the given offset."""
        return self.__class__(self.x + dx, self.y + dy, self.z + dz)

    def shift_in_direction(self, direction: Direction3D, shift: int) -> Self:
        """Shift the position in the given direction by the given shift."""
        if direction == Direction3D.X:
            return self.shift_by(dx=shift)
        elif direction == Direction3D.Y:
            return self.shift_by(dy=shift)
        else:
            return self.shift_by(dz=shift)

    def is_neighbour(self, other: Position3D) -> bool:
        """Check if the other position is near to this position, i.e. Manhattan distance is 1."""
        return abs(self.x - other.x) + abs(self.y - other.y) + abs(self.z - other.z) == 1

    def as_tuple(self) -> tuple[int, int, int]:
        """Return the position as a tuple."""
        return astuple(self)

    def __str__(self) -> str:
        return f"({self.x},{self.y},{self.z})"

    def as_2d(self) -> Position2D:
        """Return the position as a 2D position."""
        return Position2D(self.x, self.y)


class BlockPosition3D(Position3D):
    """Represent the position of a block in 3D space."""

    def as_2d(self) -> BlockPosition2D:
        """Return ``self`` as a 2-dimensional position, ignoring the ``z`` coordinate."""
        return BlockPosition2D(self.x, self.y)


class Direction3D(Enum):
    """Axis directions in the 3D spacetime diagram."""

    X = 0
    Y = 1
    Z = 2

    @staticmethod
    def all_directions() -> list[Direction3D]:
        """Return all the directions."""
        return [Direction3D.X, Direction3D.Y, Direction3D.Z]

    @staticmethod
    def spatial_directions() -> list[Direction3D]:
        """Return all the spatial directions."""
        return [Direction3D.X, Direction3D.Y]

    @staticmethod
    def temporal_directions() -> list[Direction3D]:
        """Return all the temporal directions."""
        return [Direction3D.Z]

    def __str__(self) -> str:
        return self.name

    @staticmethod
    def from_neighbouring_positions(source: Position3D, sink: Position3D) -> Direction3D:
        """Return the direction to go from ``source`` to ``sink``."""
        assert source.is_neighbour(sink)
        for direction, (source_coord, sink_coord) in zip(
            Direction3D.all_directions(), zip(source.as_tuple(), sink.as_tuple())
        ):
            if source_coord != sink_coord:
                return direction
        raise TQECError(
            f"Could not find the direction from two neighbouring positions {source:=} and {sink:=}."
        )

    @property
    def orthogonal_directions(self) -> tuple[Direction3D, Direction3D]:
        """Return the two directions orthogonal to ``self``."""
        i = self.value
        return Direction3D((i + 1) % 3), Direction3D((i + 2) % 3)


@dataclass(frozen=True)
class SignedDirection3D:
    """Signed directions in the 3D spacetime diagram."""

    direction: Direction3D
    towards_positive: bool

    def __neg__(self) -> SignedDirection3D:
        return SignedDirection3D(self.direction, not self.towards_positive)

    def __str__(self) -> str:
        return f"{'+' if self.towards_positive else '-'}{self.direction}"

    @staticmethod
    def from_string(s: str) -> SignedDirection3D:
        """Return the signed direction from a string.

        Args:
            s: The string representation of the signed direction. The string
                should have the format "<sign><direction>", where "<direction>"
                is one of "X", "Y", "Z" and "<sign>" is either "+" or "-".

        Returns:
            The signed direction.

        Raises:
            TQECError: If the string does not match the expected format.

        """
        match = re.match(r"([+-])([XYZ])", s)
        if match is None:
            raise TQECError(f"Invalid signed direction: {s}, expected format: [+/-][XYZ]")
        sign, direction = match.groups()
        return SignedDirection3D(Direction3D("XYZ".index(direction)), sign == "+")


@dataclass(frozen=True, order=True)
class FloatPosition3D:
    """A 3D float position."""

    x: float
    y: float
    z: float

    def shift_by(self, dx: float = 0, dy: float = 0, dz: float = 0) -> FloatPosition3D:
        """Shift the position by the given offset."""
        return FloatPosition3D(self.x + dx, self.y + dy, self.z + dz)

    def shift_in_direction(self, direction: Direction3D, shift: float) -> FloatPosition3D:
        """Shift the position in the given direction by the given shift."""
        if direction == Direction3D.X:
            return self.shift_by(dx=shift)
        elif direction == Direction3D.Y:
            return self.shift_by(dy=shift)
        else:
            return self.shift_by(dz=shift)

    def as_array(self) -> npt.NDArray[np.float32]:
        """Return the position as a numpy array."""
        return np.asarray(astuple(self), dtype=np.float32)



from __future__ import annotations

import hashlib
import json
import pickle
import warnings
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, ClassVar, Final

import numpy
import semver

from tqec.circuit.measurement_map import MeasurementRecordsMap
from tqec.circuit.moment import Moment
from tqec.circuit.schedule import (
    Schedule,
    ScheduledCircuit,
    relabel_circuits_qubit_indices,
)
from tqec.compile.detectors.detector import Detector
from tqec.compile.generation import generate_circuit_from_instantiation
from tqec.plaquette.plaquette import Plaquette, Plaquettes
from tqec.templates.subtemplates import SubTemplateType
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Shift2D

CURRENT_DATABASE_VERSION: Final[semver.Version] = semver.Version(1, 0, 0)


@dataclass(frozen=True)
class _DetectorDatabaseKey:
    """Immutable type used as a key in the database of detectors.

    This class represents a "situation" for which we might be able to compute
    several detectors. Its purpose of existence is to provide sensible
    `__hash__` and `__eq__` operations in order to be able to use a "situation"
    as a `dict` key.

    Attributes:
        subtemplates: a sequence of 2-dimensional arrays of integers representing
            the sub-template(s). Each entry corresponds to one QEC round.
        plaquettes_by_timestep: a list of :class:`Plaquettes`, each
            :class:`Plaquettes` entry storing enough :class:`Plaquette`
            instances to generate a circuit from corresponding entry in
            `self.subtemplates` and corresponding to one QEC round.

    ## Implementation details

    This class uses a surjective representation to compare (`__eq__`) and hash
    (`__hash__`) its instances. This representation is computed and cached using
    the :meth:`_DetectorDatabaseKey.plaquette_names` property that basically
    uses the provided subtemplates to build a nested tuple data-structure with
    the same shape as `self.subtemplates` (3 dimensions, the first one being the
    number of time steps, the next 2 ones being of odd and equal size and
    depending on the radius used to build subtemplates) storing in each of its
    entries the corresponding plaquette name.

    This intermediate data-structure is not the most memory efficient one, but
    it has the advantage of being easy to construct, trivially invariant to
    plaquette re-indexing and easy to hash (with some care to NOT use Python's
    default `hash` due to its absence of stability across different runs).

    """

    subtemplates: Sequence[SubTemplateType]
    plaquettes_by_timestep: Sequence[Plaquettes]

    def __post_init__(self) -> None:
        if len(self.subtemplates) != len(self.plaquettes_by_timestep):
            raise TQECError(
                "DetectorDatabaseKey can only store an equal number of "
                f"subtemplates and plaquettes. Got {len(self.subtemplates)} "
                f"subtemplates and {len(self.plaquettes_by_timestep)} plaquettes."
            )

    @property
    def num_timeslices(self) -> int:
        return len(self.subtemplates)

    @cached_property
    def plaquette_names(self) -> tuple[tuple[tuple[str, ...], ...], ...]:
        """Cached property returning a representation of the current situation with plaquette names.

        Returns:
            Nested tuples such that
            ``ret[t][y][x] == self.plaquettes_by_timestep[t][self.subtemplates[t][y, x]].name``.

            The returned object can be iterated on using:

            .. code:: python

                for t, names in enumerate(self.plaquette_names):
                    subtemplate = self.subtemplates[t]
                    plaquettes = self.plaquettes_by_timestep[t]
                    for y, names_row in enumerate(names):
                        for x, name in enumerate(names_row):
                            plaquette: Plaquette = plaquettes[subtemplate[y, x]]
                            assert name == plaquette.name

        """
        return tuple(
            tuple(tuple(plaquettes[pi].name for pi in row) for row in st)
            for st, plaquettes in zip(self.subtemplates, self.plaquettes_by_timestep)
        )

    @cached_property
    def reliable_hash(self) -> int:
        """Return a hash of ``self`` that is guaranteed to be constant.

        Python's ``hash`` is not guaranteed to be constant across Python versions, OSes and
        executions. In particular, strings hash will not be repeatable across different Python
        executable calls. The following line should return different values at each call:

        .. code:: bash

            python -c 'print(hash("Hello world"))'

        For example, here are the results obtained after calling that line 3 times on my machine:

        - ``2656127643635015930``
        - ``1413792191058799258``
        - ``8731178165517315210``

        This is an issue for the detector database as we would like it to be reproducible. Else,
        reusing an existing database will always fail because keys will have a different hash,
        hence re-computing detectors at each call and growing the database indefinitely.

        This method implements a reliable hash that should be constant no matter the context
        (different Python calls, different OS, different version of Python, ...).
        """
        hasher = hashlib.md5()
        for timeslice in self.plaquette_names:
            for row in timeslice:
                for name in row:
                    hasher.update(name.encode())
        return int(hasher.hexdigest(), 16)

    def __hash__(self) -> int:
        return self.reliable_hash

    def __eq__(self, rhs: object) -> bool:
        return isinstance(rhs, _DetectorDatabaseKey) and self.plaquette_names == rhs.plaquette_names

    def circuit(self, plaquette_increments: Shift2D) -> ScheduledCircuit:
        """Get the `stim.Circuit` instance represented by `self`.

        Args:
            plaquette_increments: displacement between each plaquette origin.

        Returns:
            `stim.Circuit` instance represented by `self`.

        """
        circuits, qubit_map = relabel_circuits_qubit_indices(
            [
                generate_circuit_from_instantiation(subtemplate, plaquettes, plaquette_increments)
                for subtemplate, plaquettes in zip(self.subtemplates, self.plaquettes_by_timestep)
            ]
        )
        moments: list[Moment] = list(circuits[0].moments)
        schedule: Schedule = circuits[0].schedule
        for circuit in circuits[1:]:
            moments.extend(circuit.moments)
            schedule.append_schedule(circuit.schedule)
        return ScheduledCircuit(moments, schedule, qubit_map)

    def to_dict(self, plaquettes_to_indices: dict[Plaquette, int]) -> dict[str, Any]:
        """Return a dictionary representation of the key.

        Args:
            plaquettes_to_indices: mapping from each :class:`Plaquette` to its
                index in the list of unique plaquettes. Each plaquette is
                represented by its index in the list of unique
                plaquettes to save space.

        Returns:
            a dictionary with the keys ``subtemplates`` and
            ``plaquettes_by_timestep`` and their corresponding values.

        """
        return {
            "subtemplates": [st.tolist() for st in self.subtemplates],
            "plaquettes_by_timestep": [
                p.to_dict(plaquettes_to_indices) for p in self.plaquettes_by_timestep
            ],
        }

    @staticmethod
    def from_dict(
        data: dict[str, Any],
        plaquettes: Sequence[Plaquette],
    ) -> _DetectorDatabaseKey:
        """Return a key from its dictionary representation.

        Args:
            data: dictionary with the keys ``subtemplates`` and
                ``plaquettes_by_timestep``.
            plaquettes: list of :class:`Plaquette` instances to use to build the
                :class:`Plaquettes` instances. Each plaquette is represented by
                its index in the list of unique plaquettes to save space.

        Returns:
            a new instance of :class:`_DetectorDatabaseKey` with the provided
            ``subtemplates`` and ``plaquettes_by_timestep``.

        """
        subtemplates = [numpy.array(st) for st in data["subtemplates"]]
        plaquettes_by_timestep = [
            Plaquettes.from_dict(p, plaquettes) for p in data["plaquettes_by_timestep"]
        ]
        return _DetectorDatabaseKey(subtemplates, plaquettes_by_timestep)


class _DetectorDatabaseIO:
    @staticmethod
    def _handle_load_error(filepath: Path, exception: Exception, ext: str) -> DetectorDatabase:
        moving_location = filepath.parent / f"faulty_database_{hash(exception)}.{ext}"
        warnings.warn(
            f"Error ({type(exception).__name__}) "
            f"when reading the database at {filepath}: {exception}. "
            f"Moving the database to {moving_location} and returning an empty database."
        )
        filepath.rename(moving_location)
        return DetectorDatabase()

    @staticmethod
    def from_pickle_file(filepath: Path) -> DetectorDatabase:
        try:
            with open(filepath, "rb") as f:
                database = pickle.load(f)
        except Exception as e:
            return _DetectorDatabaseIO._handle_load_error(filepath, e, "pkl")

        if not isinstance(database, DetectorDatabase):
            raise TQECError(
                f"Found the Python type {type(database).__name__} in the "
                f"provided file but {type(DetectorDatabase).__name__} was "
                "expected."
            )
        return database

    @staticmethod
    def from_json_file(filepath: Path) -> DetectorDatabase:
        try:
            with open(filepath) as f:
                data = json.load(f)
                database = DetectorDatabase.from_dict(data)
        except Exception as e:
            return _DetectorDatabaseIO._handle_load_error(filepath, e, "json")

        if not isinstance(database, DetectorDatabase):
            raise TQECError(
                f"Found the Python type {type(database).__name__} in the "
                f"provided file but {type(DetectorDatabase).__name__} was "
                "expected."
            )
        return database

    @staticmethod
    def to_pickle_file(filepath: Path, database: DetectorDatabase) -> None:
        with open(filepath, "wb") as f:
            pickle.dump(database, f)

    @staticmethod
    def to_json_file(filepath: Path, database: DetectorDatabase) -> None:
        with open(filepath, "w") as f:
            json.dump(database.to_dict(), f)


def _get_database_format(filepath: Path) -> str:
    suffix = filepath.suffix.lower()
    if suffix in {".pkl", ".pickle"}:
        return "pickle"
    if suffix in {".json"}:
        return "json"
    raise TQECError(
        f"Could not infer the database format from the provided filepath ('{filepath}'). "
        "Supported formats are:\n  -" + "\n  -".join(DetectorDatabase._WRITERS.keys())
    )


class DetectorDatabase:
    version: semver.Version = semver.Version(0, 0, 0)

    _READERS: ClassVar[Mapping[str, Callable[[Path], DetectorDatabase]]] = {
        "pickle": _DetectorDatabaseIO.from_pickle_file,
        "json": _DetectorDatabaseIO.from_json_file,
    }
    _WRITERS: ClassVar[Mapping[str, Callable[[Path, DetectorDatabase], None]]] = {
        "pickle": _DetectorDatabaseIO.to_pickle_file,
        "json": _DetectorDatabaseIO.to_json_file,
    }

    def __init__(
        self,
        mapping: dict[_DetectorDatabaseKey, frozenset[Detector]] | None = None,
        frozen: bool = False,
    ):
        """Store a mapping from "situations" to the corresponding detectors.

        This class aims at storing efficiently a set of "situations" in which the
        corresponding detectors are known and do not have to be re-computed.

        In this class, a "situation" is described by :class:`_DetectorDatabaseKey`
        and correspond to a spatially and temporally local piece of a larger
        computation.

        The version number should be manually updated when code is pushed which makes old
        instances of the database incompatible with newly generated instances.
        Guidance on when to change `a` (major) or `b` (minor) in the `a.b` version number:
        - MAJOR when the format of the file changes (i.e. when the attributes of
        ``DetectorDatabase`` change),
        - MINOR when the content of the database is invalidated (e.g. by changing a plaquette
        implementation without changing its name).

        Old databases generated prior to the introduction of a version attribute will be
        loaded with the default value of .version, without passing through __init__,
        ie (0,0,0).

        """
        if mapping is None:
            mapping = dict()
        self.mapping = mapping
        self.frozen = frozen
        self.version = CURRENT_DATABASE_VERSION

    def add_situation(
        self,
        subtemplates: Sequence[SubTemplateType],
        plaquettes_by_timestep: Sequence[Plaquettes],
        detectors: frozenset[Detector] | Detector,
    ) -> None:
        """Add a new situation to the database.

        Args:
            subtemplates: a sequence of 2-dimensional arrays of integers
                representing the sub-template(s). Each entry corresponds to one
                QEC round.
            plaquettes_by_timestep: a list of :class:`Plaquettes`, each
                :class:`Plaquettes` entry storing enough :class:`Plaquette`
                instances to generate a circuit from corresponding entry in
                `self.subtemplates` and corresponding to one QEC round.
            detectors: computed detectors that should be stored in the database.
                The coordinates used by the :class:`Measurement` instances stored
                in each entry should be relative to the top-left qubit of the
                top-left plaquette in the provided `subtemplates`.

        Raises:
            TQECError: if this method is called and `self.frozen`.

        """
        if self.frozen:
            raise TQECError("Cannot add a situation to a frozen database.")
        key = _DetectorDatabaseKey(subtemplates, plaquettes_by_timestep)
        self.mapping[key] = frozenset([detectors]) if isinstance(detectors, Detector) else detectors

    def remove_situation(
        self,
        subtemplates: Sequence[SubTemplateType],
        plaquettes_by_timestep: Sequence[Plaquettes],
    ) -> None:
        """Remove an existing situation from the database.

        Args:
            subtemplates: a sequence of 2-dimensional arrays of integers
                representing the sub-template(s). Each entry corresponds to one
                QEC round.
            plaquettes_by_timestep: a list of :class:`Plaquettes`, each
                :class:`Plaquettes` entry storing enough :class:`Plaquette`
                instances to generate a circuit from corresponding entry in
                `self.subtemplates` and corresponding to one QEC round.

        Raises:
            TQECError: if this method is called and `self.frozen`.

        """
        if self.frozen:
            raise TQECError("Cannot remove a situation to a frozen database.")
        key = _DetectorDatabaseKey(subtemplates, plaquettes_by_timestep)
        del self.mapping[key]

    def get_detectors(
        self,
        subtemplates: Sequence[SubTemplateType],
        plaquettes_by_timestep: Sequence[Plaquettes],
    ) -> frozenset[Detector] | None:
        """Return the detectors associated with the provided situation.

        Args:
            subtemplates: a sequence of 2-dimensional arrays of integers
                representing the sub-template(s). Each entry corresponds to one
                QEC round.
            plaquettes_by_timestep: a list of :class:`Plaquettes`, each
                :class:`Plaquettes` entry storing enough :class:`Plaquette`
                instances to generate a circuit from corresponding entry in
                `self.subtemplates` and corresponding to one QEC round.

        Returns:
            detectors associated with the provided situation or `None` if the
            situation is not in the database.

        """
        key = _DetectorDatabaseKey(subtemplates, plaquettes_by_timestep)
        return self.mapping.get(key)

    def freeze(self) -> None:
        """Make ``self`` read-only."""
        self.frozen = True

    def unfreeze(self) -> None:
        """Make ``self`` writable."""
        self.frozen = False

    def to_crumble_urls(self, plaquette_increments: Shift2D = Shift2D(2, 2)) -> list[str]:
        """Return a URL pointing to https://algassert.com/crumble for each of the stored situations.

        Args:
            plaquette_increments: increments between two :class:`Plaquette`
                origins. Default to `Displacement(2, 2)` which is the expected
                value for surface code.

        Returns:
            a list of Crumble URLs, each one representing a situation stored in
            `self`.

        """
        urls: list[str] = []
        for key, detectors in self.mapping.items():
            circuit = key.circuit(plaquette_increments)
            rec_map = MeasurementRecordsMap.from_scheduled_circuit(circuit)
            for detector in detectors:
                circuit.append_annotation(detector.to_instruction(rec_map))
            urls.append(circuit.get_circuit().to_crumble_url())
        return urls

    def __len__(self) -> int:
        return len(self.mapping)

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the database.

        Returns:
            a dictionary with the keys ``mapping`` and ``frozen`` and their
            corresponding values.

        """
        # First obtain the unique plaquettes
        plaquettes_set: set[Plaquette] = set()
        for key in self.mapping:
            for p in key.plaquettes_by_timestep:
                if p.collection.default_value is not None:
                    plaquettes_set.add(p.collection.default_value)
                plaquettes_set.update(p.collection.values())
        uniq_plaquettes: Sequence[Plaquette] = list(plaquettes_set)
        # Then create a mapping from each plaquette to its index
        plaquettes_to_indices = {p: i for i, p in enumerate(uniq_plaquettes)}
        return {
            "mapping": [
                [
                    key.to_dict(plaquettes_to_indices=plaquettes_to_indices),
                    [d.to_dict() for d in detectors],
                ]
                for key, detectors in self.mapping.items()
            ],
            "frozen": self.frozen,
            "uniq_plaquettes": [p.to_dict() for p in uniq_plaquettes],
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> DetectorDatabase:
        """Return a database from its dictionary representation.

        Args:
            data: dictionary with the keys ``mapping`` and ``frozen``.

        Returns:
            a new instance of :class:`DetectorDatabase` with the provided
            ``mapping`` and ``frozen``.

        """
        uniq_plaquettes = [Plaquette.from_dict(p) for p in data["uniq_plaquettes"]]
        mapping = {
            _DetectorDatabaseKey.from_dict(key, plaquettes=uniq_plaquettes): frozenset(
                Detector.from_dict(d) for d in detectors
            )
            for key, detectors in data["mapping"]
        }
        return DetectorDatabase(mapping, data["frozen"])

    def to_file(self, filepath: Path) -> None:
        """Save the database to a file.

        Args:
            filepath: path to the file where the database should be saved.

        """
        if not filepath.parent.exists():
            filepath.parent.mkdir(parents=True)
        format = _get_database_format(filepath)
        DetectorDatabase._WRITERS[format](filepath, self)

    @staticmethod
    def from_file(filepath: Path) -> DetectorDatabase:
        """Initialise a new instance from a file.

        Args:
            filepath: path to a file where a :class:`.DetectorDatabase` instance has been saved.

        Returns:
            a new :class:`.DetectorDatabase` instance read from the provided ``filepath``.

        """
        if not filepath.exists():
            raise TQECError(
                f"Could not read the database: the provided filepath ('{filepath}') does not exist "
                "on disk."
            )
        format = _get_database_format(filepath)
        database = DetectorDatabase._READERS[format](filepath)
        return database



"""Defines several scalable classes such as :class:`LinearFunction` or :class:`Scalable2D`.

This module defines the necessary classes to help with scalable structures.
:class:`LinearFunction` simply represents a linear function ``a * k + b`` where
``a`` and ``b`` are floating-point quantities.

This :class:`LinearFunction` class is for example used to represent the shape
of a :class:`~tqec.templates.base.Template` for any input value ``k``.
More specifically, :class:`~tqec.templates.qubit.QubitTemplate` has a
shape that should exactly match a pair of ``LinearFunction(2, 2)`` which is
basically ``2k + 2``.

:class:`Scalable2D` is exactly made to represent such pairs of scalable quantities.

"""

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from typing import cast

from typing_extensions import Self

from tqec.circuit.qubit import GridQubit
from tqec.utils.exceptions import TQECError
from tqec.utils.position import PhysicalQubitShape2D, PlaquetteShape2D, Shape2D, Shift2D


@dataclass(frozen=True)
class LinearFunction:
    """Represent a linear function.

    A linear function is fully described with a slope and an offset.

    """

    slope: float = 1.0
    offset: float = 0.0

    def __call__(self, x: float) -> float:
        """Evaluate the linear function on a given input.

        Args:
            x: the input to evaluate the linear function on.

        Returns:
            the image of x.

        """
        return self.slope * x + self.offset

    def __add__(self, other: LinearFunction | int | float) -> LinearFunction:
        """Add two linear functions and returns the result.

        This method does not modify self in-place.

        Args:
            other: the right-hand side to add to self.

        Returns:
            a new linear function instance representing `self + other`.

        """
        if isinstance(other, (int, float)):
            other = LinearFunction(0, other)
        return LinearFunction(self.slope + other.slope, self.offset + other.offset)

    def __sub__(self, other: LinearFunction | int | float) -> LinearFunction:
        """Subtract two linear functions and returns the result.

        This method does not modify self in-place.

        Args:
            other: the right-hand side to subtract to self.

        Returns:
            a new linear function instance representing `self - other`.

        """
        if isinstance(other, (int, float)):
            other = LinearFunction(0, other)
        return LinearFunction(self.slope - other.slope, self.offset - other.offset)

    def __mul__(self, other: int | float) -> LinearFunction:
        """Multiply a linear function by a scalar.

        Args:
            other: the scalar that should multiply self.

        Returns:
            a copy of `self`, scaled by the provided `other`.

        """
        return self.__rmul__(other)

    def __rmul__(self, other: int | float) -> LinearFunction:
        """Multiply a linear function by a scalar.

        Args:
            other: the scalar that should multiply self.

        Returns:
            a copy of `self`, scaled by the provided `other`.

        """
        return LinearFunction(other * self.slope, other * self.offset)

    def intersection(self, other: LinearFunction) -> float | None:
        """Compute the intersection between two linear functions.

        Args:
            other: the `LinearFunction` instance to intersect with `self`.

        Returns:
            If they intersect, return x such that `self(x) = other(x)`.
            Otherwise, return None.

        """
        if self.slope == other.slope:
            return None

        return -(other.offset - self.offset) / (other.slope - self.slope)

    @staticmethod
    def _from(obj: LinearFunction | float) -> LinearFunction:  # pragma: no cover
        if isinstance(obj, (float, int)):
            return LinearFunction(0, obj)
        else:
            return obj

    def __repr__(self) -> str:
        if abs(self.slope) < 1e-8:
            return str(self.offset)
        if abs(self.offset) < 1e-8:
            return f"{self.slope}*x"
        return f"{self.slope}*x + {self.offset}"

    def integer_eval(self, x: int) -> int:
        """Evaluate the linear function on ``x`` and return the result as an integer.

        Raises:
            TQECError: if the result is not an integer.

        Returns:
            ``int(self(x))``.

        """
        return round_or_fail(self.slope * x + self.offset)

    def exact_integer_div(self, div: int) -> LinearFunction:
        """Divide ``self`` by ``div`` when possible.

        Raises:
            ZeroDivisionError: if ``div == 0``.
            TQECError: if ``self.slope`` or ``self.offset`` are not divisible by ``div``.

        Returns:
            a new linear function such that ``div * ret == self``.

        """
        if div == 0:
            raise ZeroDivisionError()
        slope, offset = round_or_fail(self.slope), round_or_fail(self.offset)
        if slope % div != 0:
            raise TQECError(
                "Trying to divide exactly a LinearFunction by an integer that "
                f"is not a multiple of the slope. Divisor: {div}. Slope: {slope}."
            )
        if offset % div != 0:
            raise TQECError(
                "Trying to divide exactly a LinearFunction by an integer that "
                f"is not a multiple of the offset. Divisor: {div}. Offset: "
                f"{offset}."
            )
        return LinearFunction(slope // div, offset // div)

    def is_constant(self, atol: float = 1e-8) -> bool:
        """Return ``True`` if ``self.slope`` is close to ``0``."""
        return abs(self.slope) < atol

    def is_scalable(self, atol: float = 1e-8) -> bool:
        """Return ``True`` if ``not self.is_scalable``."""
        return not self.is_constant(atol)

    def is_close_to(self, other: LinearFunction, atol: float = 1e-8) -> bool:
        """Return ``True`` is ``self`` is approximately equal to ``other``."""
        return abs(self.slope - other.slope) < atol and abs(self.offset - other.offset) < atol

    @staticmethod
    def unambiguous_max_on_positives(
        fs: Iterable[LinearFunction], default: LinearFunction | None = None
    ) -> LinearFunction:
        """Compute the unambiguous maximum of the provided linear functions on the positive numbers.

        A unambiguous maximum on R+ (the set of positive numbers) is a linear
        function that is greater or equal than all the functions in ``fs`` on
        the whole R+ interval.

        Args:
            fs: linear functions to find a unambiguous maximum in.
            default: default value to return if ``fs`` is empty. Defaults to
                ``None`` which is internally translated to ``LinearFunction(0, 0)``.

        Raises:
            TQECError: if the maximum found is ambiguous.

        Returns:
            the unambiguous maximum in ``fs``.

        """
        if default is None:
            default = LinearFunction(0, 0)
        iterator = iter(fs)
        try:
            res: LinearFunction = next(iterator)
        except StopIteration:
            return default
        # Find the **potentially ambiguous** maximum trivially.
        # If the maximum is unambiguous, res will be it. Else, there is an
        # ambiguity and so we should raise.
        for f in iterator:
            if f.offset >= res.offset and f.slope >= res.slope:
                res = f

        for f in fs:
            if f.offset > res.offset or f.slope > res.slope:
                raise TQECError(
                    "Could not find a unambiguous maximum in the provided linear functions. "
                    f"{res} could be the maximum, but is ambiguous with {f}."
                )
        return res

    @staticmethod
    def safe_mul(lhs: LinearFunction, rhs: LinearFunction) -> LinearFunction:
        """Return ``lhs * rhs``, checking that the result is a linear function.

        Raises:
            TQECError: if both ``lhs.slope`` and ``rhs.slope`` are non-zero.

        """
        if lhs.slope != 0 and rhs.slope != 0:
            raise TQECError(f"The result of ({lhs}) * ({rhs}) is not a linear function.")
        return LinearFunction(
            lhs.slope * rhs.offset + rhs.slope * lhs.offset, lhs.offset * rhs.offset
        )


def round_or_fail(f: float, atol: float = 1e-8) -> int:
    """Try to round the provided ``f`` to the nearest integer.

    Args:
        f: a floating-point value that should be close (absolute tolerance of
            ``atol``) to its nearest integer number.
        atol: absolute tolerance between the provided ``f`` and ``round(f)`` that
            is acceptable.

    Raises:
        TQECError: if ``abs(f - round(f)) > atol``

    Returns:
        ``int(round(f))``

    """
    rounded_value = round(f)
    if abs(f - rounded_value) > atol:
        raise TQECError(f"Rounding from {f} to integer failed.")
    return rounded_value


@dataclass(frozen=True)
class Scalable2D:
    """A pair of scalable quantities.

    Attributes:
        x: a linear function representing the value of the ``x`` coordinate.
        y: a linear function representing the value of the ``y`` coordinate.

    """

    x: LinearFunction
    y: LinearFunction

    def to_shape_2d(self, k: int) -> Shape2D:
        """Get the represented value for a given scaling parameter ``k``.

        Args:
            k: scaling parameter to use to get a value from the scalable
                quantities stored in ``self``.

        Raises:
            TQECError: if any of ``self.x(k)`` or ``self.y(k)`` returns a
                number that is not an integer (or very close to an integer).

        Returns:
            ``Shape2D(self.x.integer_eval(k), self.y.integer_eval(k))``

        """
        return Shape2D(self.x.integer_eval(k), self.y.integer_eval(k))

    def to_numpy_shape(self, k: int) -> tuple[int, int]:
        """Get a tuple of coordinates in ``numpy``-coordinates.

        Raises:
            TQECError: if any of ``self.x(k)`` or ``self.y(k)`` returns a
                number that is not an integer (or very close to an integer).

        Returns:
            a tuple of coordinates in ``numpy``-coordinates.

        """
        return self.to_shape_2d(k).to_numpy_shape()

    @staticmethod
    def _get_x_y(
        other: Scalable2D | Shift2D | tuple[LinearFunction | int, LinearFunction | int],
    ) -> tuple[LinearFunction | int, LinearFunction | int]:
        if isinstance(other, tuple):
            return cast(tuple[LinearFunction | int, LinearFunction | int], other)
        elif isinstance(other, (Scalable2D | Shift2D)):
            return other.x, other.y
        else:
            # added because flagged by ty
            raise TypeError("Unsupported input provided.")

    def __add__(self: Self, other: Self | Shift2D | tuple[int, int]) -> Self:
        x, y = Scalable2D._get_x_y(other)
        return self.__class__(self.x + x, self.y + y)

    def __sub__(self: Self, other: Self | Shift2D | tuple[int, int]) -> Self:
        x, y = Scalable2D._get_x_y(other)
        return self.__class__(self.x - x, self.y - y)


class PlaquetteScalable2D(Scalable2D):
    """A pair of scalable quantities in plaquette coordinates."""

    def to_shape_2d(self, k: int) -> PlaquetteShape2D:
        """Evaluate both coordinates with ``k`` as input."""
        return PlaquetteShape2D(self.x.integer_eval(k), self.y.integer_eval(k))

    def __mul__(self, other: Shift2D) -> PhysicalQubitScalable2D:
        return PhysicalQubitScalable2D(self.x * other.x, self.y * other.y)


class PhysicalQubitScalable2D(Scalable2D):
    """A pair of scalable quantities in physical qubit coordinates."""

    def to_shape_2d(self, k: int) -> PhysicalQubitShape2D:
        """Evaluate both coordinates with ``k`` as input."""
        return PhysicalQubitShape2D(self.x.integer_eval(k), self.y.integer_eval(k))

    def to_grid_qubit(self, k: int) -> GridQubit:
        """Evaluate both coordinates with ``k`` as input, returning a :class:`.GridQubit`."""
        return GridQubit(self.x.integer_eval(k), self.y.integer_eval(k))



"""Provides code to compile a block graph instance into a ``stim.Circuit``.

This module defines the needed classes and functions to transform a
:class:`~tqec.computation.block_graph.BlockGraph` instance into a fully annotated
``stim.Circuit`` instance that can be simulated using ``stim`` and even executed
on available hardware.

"""

from .compile import compile_block_graph as compile_block_graph



"""Defines a few core data-structures that are independent of other ``tqec`` modules.

The goal of this module is to host data-structures that do not clearly belong to
another existing ``tqec`` sub-module and that also do not import any code from
the other ``tqec`` sub-modules.

"""

from .enums import Basis as Basis
from .enums import Orientation as Orientation
from .exceptions import TQECError as TQECError
from .noise_model import NoiseModel as NoiseModel



from .fixed_boundary import FIXED_BOUNDARY_CUBE_BUILDER as FIXED_BOUNDARY_CUBE_BUILDER
from .fixed_boundary import FIXED_BOUNDARY_PIPE_BUILDER as FIXED_BOUNDARY_PIPE_BUILDER
from .fixed_bulk import FIXED_BULK_CUBE_BUILDER as FIXED_BULK_CUBE_BUILDER
from .fixed_bulk import FIXED_BULK_PIPE_BUILDER as FIXED_BULK_PIPE_BUILDER



import multiprocessing
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path

import sinter

from tqec.compile.compile import _DEFAULT_BLOCK_REPETITIONS, compile_block_graph
from tqec.compile.convention import FIXED_BULK_CONVENTION, Convention
from tqec.compile.detectors.database import DetectorDatabase
from tqec.computation.block_graph import BlockGraph
from tqec.computation.correlation import CorrelationSurface
from tqec.simulation.generation import generate_sinter_tasks
from tqec.simulation.split import (
    heuristic_custom_error_key,
    split_stats_for_observables,
)
from tqec.utils.noise_model import NoiseModel
from tqec.utils.paths import DEFAULT_DETECTOR_DATABASE_PATH
from tqec.utils.scale import LinearFunction


def start_simulation_using_sinter(
    block_graph: BlockGraph,
    ks: Sequence[int],
    ps: Sequence[float],
    noise_model_factory: Callable[[float], NoiseModel],
    manhattan_radius: int,
    convention: Convention = FIXED_BULK_CONVENTION,
    observables: list[CorrelationSurface] | None = None,
    detector_database: DetectorDatabase | None = None,
    database_path: str | Path = DEFAULT_DETECTOR_DATABASE_PATH,
    num_workers: int = multiprocessing.cpu_count(),
    progress_callback: Callable[[sinter.Progress], None] | None = None,
    max_shots: int | None = None,
    max_errors: int | None = None,
    decoders: Iterable[str] = ("pymatching",),
    print_progress: bool = False,
    custom_decoders: dict[str, sinter.Decoder | sinter.Sampler] | None = None,
    save_resume_filepath: str | Path | None = None,
    existing_data_filepaths: Iterable[str | Path] = (),
    split_observable_stats: bool = True,
    block_temporal_height: LinearFunction = _DEFAULT_BLOCK_REPETITIONS,
) -> list[list[sinter.TaskStats]]:
    """Run `stim` simulations using `sinter`.

    This function is the preferred entry-point to run `sinter` computations using
    `stim` from circuits generated by `tqec`.

    It removes the need to generate the `stim.Circuit` instances in parallel (due
    to the relative inefficiency of detector annotation at the moment) by
    importing and calling :func:`generate_sinter_tasks`. It also forwards
    several parameters to :func:`sinter.collect`, but without showing in its
    signature arguments that we do not envision using.

    Args:
        block_graph: a representation of the QEC computation to simulate.
        ks: values of the scaling parameter `k` to use in order to generate the
            circuits.
        ps: values of the noise parameter `p` to use to instantiate a noise
            model using the provided `noise_model_factory`.
        noise_model_factory: a callable that is used to instantiate a noise
            model from each of the noise parameters in `ps`.
        manhattan_radius: radius used to automatically compute detectors. The
            best value to set this argument to is the minimum integer such that
            flows generated from any given reset/measurement, from any plaquette
            at any spatial/temporal place in the QEC computation, do not
            propagate outside of the qubits used by plaquettes spatially located
            at maximum `manhattan_radius` plaquettes from the plaquette the
            reset/measurement belongs to (w.r.t. the Manhattan distance).
            Default to 2, which is sufficient for regular surface code. If
            negative, detectors are not computed automatically and are not added
            to the generated circuits.
        convention: convention used to generate the quantum circuits.
        observables: a list of correlation surfaces to compile to logical
             observables and generate statistics for. If `None`, all the correlation
             surfaces of the provided computation are used.
        detector_database: an instance to retrieve from / store in detectors
            that are computed as part of the circuit generation.
        database_path: str or Path where detector database is presaved
        num_workers: The number of worker processes to use.
        progress_callback: Defaults to None (unused). If specified, then each
            time new sample statistics are acquired from a worker this method
            will be invoked with the new `sinter.TaskStats`.
        max_shots: Defaults to None (unused). Stops the sampling process
            after this many samples have been taken from the circuit.
        max_errors: Defaults to None (unused). Stops the sampling process
            after this many errors have been seen in samples taken from the
            circuit. The actual number sampled errors may be larger due to
            batching.
        decoders: Defaults to None (specified by each Task). The names of the
            decoders to use on each Task. It must either be the case that each
            Task specifies a decoder and this is set to None, or this is an
            iterable and each Task has its decoder set to None.
        print_progress: When True, progress is printed to stderr while
            collection runs.
        custom_decoders: Named child classes of `sinter.decoder`, that can be
            used if requested by name by a task or by the decoders list.
            If not specified, only decoders with support built into sinter, such
            as 'pymatching' and 'fusion_blossom', can be used.
        save_resume_filepath: Defaults to None (unused). If set to a filepath,
            results will be saved to that file while they are collected. If the
            python interpreter is stopped or killed, calling this method again
            with the same save_resume_filepath will load the previous results
            from the file so it can resume where it left off. The stats in this
            file will be counted in addition to each task's previous_stats field
            (as opposed to overriding the field). Notes that the stats for each
            individual observable are not split but saved in the `custom_counts`
            field of the stats.
        existing_data_filepaths: CSV data saved to these files will be loaded,
            included in the returned results, and count towards things like
            max_shots and max_errors.
        split_observable_stats: Defaults to True. If True, the results are
            post-processed to get individual statistics for each observable in
            `observables`. If False, the results are returned as they are
            collected.
        block_temporal_height: the number of rounds of stabilizer measurements
            (ignoring one layer for initialization and another for final measurement).
            Defaults to `2k-1`.

    Returns:
        A list of lists of `sinter.TaskStats`. If `split_observable_stats` is
        True, the outer list has one element per provided observable in `observables`.
        If `split_observable_stats` is False, the outer list has only one element,
        containing the raw statistics collected.

    """
    if observables is None:
        observables = block_graph.find_correlation_surfaces()
    custom_error_count_key: str | None = None
    if split_observable_stats and len(observables) > 1:
        custom_error_count_key = heuristic_custom_error_key(observables)

    compiled_graph = compile_block_graph(
        block_graph,
        convention,
        observables,
        block_temporal_height,
    )
    stats = sinter.collect(
        num_workers=num_workers,
        tasks=generate_sinter_tasks(
            compiled_graph,
            ks,
            ps,
            noise_model_factory,
            manhattan_radius,
            detector_database,
            database_path,
        ),
        existing_data_filepaths=existing_data_filepaths,
        save_resume_filepath=save_resume_filepath,
        progress_callback=progress_callback,
        max_shots=max_shots,
        max_errors=max_errors,
        decoders=decoders,
        print_progress=print_progress,
        custom_decoders=custom_decoders,
        hint_num_tasks=len(ks) * len(ps),
        count_observable_error_combos=True,
        custom_error_count_key=custom_error_count_key,
    )
    if split_observable_stats:
        return split_stats_for_observables(stats, len(observables))
    return [stats]



"""Defines :class:`~tqec.circuit.qubit.GridQubit` and helper functions to manage qubits.

This module defines a central class to represent a qubit placed on a
2-dimensional grid, :class:`GridQubit`, and a few functions to extract
qubit-related information from ``stim.Circuit`` instances.

"""

from __future__ import annotations

from collections import defaultdict
from typing import cast

import stim

from tqec.utils.coordinates import StimCoordinates
from tqec.utils.position import PhysicalQubitPosition2D, Shift2D


class GridQubit:
    def __init__(self, x: int, y: int) -> None:
        """Represent a qubit placed on a 2-dimensional grid.

        Args:
            x: first coordinate.
            y: second coordinate.

        """
        self._x = x
        self._y = y

    @property
    def x(self) -> int:
        """Coordinate of the qubit in the first spatial dimension."""
        return self._x

    @property
    def y(self) -> int:
        """Coordinate of the qubit in the second spatial dimension."""
        return self._y

    def to_qubit_coords_instruction(self, index: int) -> stim.CircuitInstruction:
        """Return the ``QUBIT_COORDS`` instruction defining ``self`` in a ``stim.Circuit``."""
        return stim.CircuitInstruction(
            "QUBIT_COORDS",
            [index],
            StimCoordinates(self.x, self.y).to_stim_coordinates(),
        )

    def __add__(self, other: GridQubit | PhysicalQubitPosition2D | Shift2D) -> GridQubit:
        return GridQubit(self.x + other.x, self.y + other.y)

    def __sub__(self, other: GridQubit | PhysicalQubitPosition2D | Shift2D) -> GridQubit:
        return GridQubit(self.x - other.x, self.y - other.y)

    def __mul__(self, other: int) -> GridQubit:
        return GridQubit(other * self.x, other * self.y)

    def __rmul__(self, other: int) -> GridQubit:
        return GridQubit(other * self.x, other * self.y)

    def __hash__(self) -> int:
        return hash((self._x, self._y))

    def __eq__(self, value: object) -> bool:
        return isinstance(value, GridQubit) and self._x == value._x and self._y == value._y

    def __lt__(self, other: GridQubit) -> bool:
        return (self._x, self._y) < (other._x, other._y)

    def __repr__(self) -> str:
        return f"GridQubit({self.x}, {self.y})"

    def __str__(self) -> str:
        return f"Q[{self.x}, {self.y}]"

    def to_dict(self) -> dict[str, int]:
        """Return a dictionary representation of the qubit.

        Returns:
            a dictionary with the keys ``x`` and ``y`` and their
            corresponding values.

        """
        return {"x": self.x, "y": self.y}

    @staticmethod
    def from_dict(data: dict[str, int]) -> GridQubit:
        """Return a qubit from its dictionary representation.

        Args:
            data: dictionary with the keys ``x`` and ``y``.

        Returns:
            a new instance of :class:`GridQubit` with the provided
            ``x`` and ``y``.

        """
        return GridQubit(data["x"], data["y"])


"""Names of the `stim` instructions that are considered as annotations."""
ANNOTATION_INSTRUCTIONS: frozenset[str] = frozenset(
    [
        # Noise channels
        "CORRELATED_ERROR",
        "DEPOLARIZE1",
        "DEPOLARIZE2",
        "E",
        "ELSE_CORRELATED_ERROR",
        "HERALDED_ERASE",
        "HERALDED_PAULI_CHANNEL_1",
        "PAULI_CHANNEL_1",
        "PAULI_CHANNEL_2",
        "X_ERROR",
        "Y_ERROR",
        "Z_ERROR",
        # Annotations
        "DETECTOR",
        "MPAD",
        "OBSERVABLE_INCLUDE",
        "QUBIT_COORDS",
        "SHIFT_COORDS",
        "TICK",
    ]
)


def count_qubit_accesses(circuit: stim.Circuit) -> dict[int, int]:
    """Count the number of times a given qubit is used by an instruction that is not an annotation.

    Note:
        If a ``REPEAT`` instruction is found, each qubit access within the
        repeated block will be multiplied by the number of time the block is
        repeated.

    Args:
        circuit: circuit containing the gates.

    Returns:
        a mapping from qubit indices (as keys) to the number of non-annotation
        instructions that have this qubit index as target (as values).

    """
    counter: defaultdict[int, int] = defaultdict(int)
    for instruction in circuit:
        if isinstance(instruction, stim.CircuitRepeatBlock):
            for qi, count in count_qubit_accesses(instruction.body_copy()).items():
                counter[qi] += count * instruction.repeat_count
        else:
            if instruction.name in ANNOTATION_INSTRUCTIONS:
                continue
            for target in instruction.targets_copy():
                # Ignore targets that are not qubit targets.
                if not target.is_qubit_target:
                    continue
                qi = cast(int, target.qubit_value)
                counter[qi] += 1
    return counter


def get_used_qubit_indices(circuit: stim.Circuit) -> set[int]:
    """Return the indices of qubits that are used by at least one non-annotation instruction.

    Args:
        circuit: circuit containing the gates.

    Returns:
        the set of qubit indices that are used by at least one non-annotation
        instruction.

    """
    return set(count_qubit_accesses(circuit).keys())



"""Implements the fixed boundary convention.

This module is intended to be the only module through which instances of :class:`.Plaquette` can be
obtained for the fixed boundary convention.

"""

from __future__ import annotations

from typing import Final, Literal

from tqec.compile.specs.base import CubeSpec
from tqec.compile.specs.enums import SpatialArms
from tqec.compile.specs.library.generators.constants import (
    HORIZONTAL_HOOK_SCHEDULES,
    VERTICAL_HOOK_SCHEDULES,
)
from tqec.compile.specs.library.generators.extended_stabilizers import ExtendedPlaquetteCollection
from tqec.compile.specs.library.generators.utils import PlaquetteMapper
from tqec.plaquette.compilation.base import PlaquetteCompiler
from tqec.plaquette.enums import PlaquetteOrientation
from tqec.plaquette.plaquette import Plaquette, Plaquettes
from tqec.plaquette.rpng.rpng import ExtendedBasis, RPNGDescription
from tqec.plaquette.rpng.translators.base import RPNGTranslator
from tqec.templates.base import RectangularTemplate
from tqec.templates.qubit import (
    QubitHorizontalBorders,
    QubitSpatialCubeTemplate,
    QubitTemplate,
    QubitVerticalBorders,
)
from tqec.utils.enums import Basis, Orientation
from tqec.utils.exceptions import TQECError
from tqec.utils.frozendefaultdict import FrozenDefaultDict
from tqec.utils.position import Direction3D


class FixedBoundaryConventionGenerator:
    def __init__(self, translator: RPNGTranslator, compiler: PlaquetteCompiler):
        """Low-level helper class centralising all the plaquette creation for the fixed boundary
        convention.

        This class provides all the methods needed to return the appropriate templates and
        plaquettes needed to implement low-level QEC layers using the fixed boundary convention.

        """
        self._mapper = PlaquetteMapper(translator, compiler)

    def get_bulk_rpng_descriptions(
        self,
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
        reset_and_measured_indices: tuple[Literal[0, 1, 2, 3], ...] = (0, 1, 2, 3),
    ) -> dict[Basis, dict[Orientation, RPNGDescription]]:
        """Get plaquettes that are supposed to be used in the bulk.

        This function returns the four 4-body stabilizer measurement plaquettes
        containing 5 rounds that can be arbitrarily tiled without any gate schedule
        clash. These plaquettes are organised by basis and hook orientation.

        Args:
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            reset: basis of the reset operation performed on data-qubits. Defaults
                to ``None`` that translates to no reset being applied on data-qubits.
            measurement: basis of the measurement operation performed on data-qubits.
                Defaults to ``None`` that translates to no measurement being applied
                on data-qubits.
            reset_and_measured_indices: data-qubit indices that should be impacted
                by the provided ``reset`` and ``measurement`` values.

        Returns:
            a mapping with 4 plaquettes: one for each basis (either ``X`` or ``Z``)
            and for each hook orientation (either ``HORIZONTAL`` or ``VERTICAL``).

        """
        # r/m: reset/measurement basis applied to each data-qubit in ``reset_and_measured_indices``
        r = reset.value.lower() if reset is not None else "-"
        m = measurement.value.lower() if measurement is not None else "-"
        # rs/ms: resets/measurements basis applied for each data-qubit
        rs = [r if i in reset_and_measured_indices else "-" for i in range(4)]
        ms = [m if i in reset_and_measured_indices else "-" for i in range(4)]
        # 2-qubit gate schedules
        vsched = VERTICAL_HOOK_SCHEDULES[is_reversed]
        hsched = HORIZONTAL_HOOK_SCHEDULES[is_reversed]
        return {
            basis: {
                Orientation.VERTICAL: RPNGDescription.from_string(
                    " ".join(f"{r}{basis.value.lower()}{s}{m}" for r, s, m in zip(rs, vsched, ms))
                ),
                Orientation.HORIZONTAL: RPNGDescription.from_string(
                    " ".join(f"{r}{basis.value.lower()}{s}{m}" for r, s, m in zip(rs, hsched, ms))
                ),
            }
            for basis in Basis
        }

    def get_3_body_rpng_descriptions(
        self,
        basis: Basis,
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> tuple[RPNGDescription, RPNGDescription, RPNGDescription, RPNGDescription]:
        """Return the four 3-body stabilizer measurement plaquettes.

        Args:
            basis: basis of the 3-body stabilizer that should be measured by the returned
                plaquettes.
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            reset: basis of the reset operation performed on data-qubits. Defaults
                to ``None`` that translates to no reset being applied on data-qubits.
            measurement: basis of the measurement operation performed on data-qubits.
                Defaults to ``None`` that translates to no measurement being applied
                on data-qubits.

        Returns:
            the four 3-body stabilizer measurement plaquettes. Their order follow the usual
            convention: ``(top_left, top_right, bottom_left, bottom_right)``.

        """
        b = basis.value.lower()
        # r/m: reset/measurement basis applied to each data-qubit
        r = reset.value.lower() if reset is not None else "-"
        m = measurement.value.lower() if measurement is not None else "-"
        # Note: the schedule of CNOT gates in corner plaquettes is less important
        # because hook errors do not exist on 3-body stabilizers. We arbitrarily
        # pick the vertical schedule.
        s = VERTICAL_HOOK_SCHEDULES[is_reversed]
        # Note that we include resets and measurements on all the used data-qubits.
        # That should be fine because this plaquette only touches cubes and pipes
        # that are related to the spatial junction being implemented, and it is not
        # valid to have a temporal pipe coming from below a spatial junction, hence
        # the data-qubits cannot be already initialised to a value we would like to
        # keep and that would be destroyed by reset/measurement.
        return (
            RPNGDescription.from_string(f"---- {r}{b}{s[1]}{m} {r}{b}{s[2]}{m} {r}{b}{s[3]}{m}"),
            RPNGDescription.from_string(f"{r}{b}{s[0]}{m} ---- {r}{b}{s[2]}{m} {r}{b}{s[3]}{m}"),
            RPNGDescription.from_string(f"{r}{b}{s[0]}{m} {r}{b}{s[1]}{m} ---- {r}{b}{s[3]}{m}"),
            RPNGDescription.from_string(f"{r}{b}{s[0]}{m} {r}{b}{s[1]}{m} {r}{b}{s[2]}{m} ----"),
        )

    def get_2_body_rpng_descriptions(
        self, is_reversed: bool, hadamard: bool = False
    ) -> dict[Basis, dict[PlaquetteOrientation, RPNGDescription]]:
        """Get plaquettes that are supposed to be used on the boundaries.

        This function returns the eight 2-body stabilizer measurement plaquettes
        that can be used on the 5-round plaquettes returned by
        :meth:`get_bulk_rpng_descriptions`.

        Args:
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            hadamard: ``True`` if the plaquette should contain a Hadamard gate.

        Note:
            The 2-body stabilizer measurement plaquettes returned by this function
            all follow the same schedule: ``1-2-3-5``.

        Warning:
            By convention, the 2-body stabilizers never reset/measure any
            data-qubit. This is done because it is way simpler to reset the correct
            data-qubits in 4-body stabilizers, and the resets/measurements in 2-body
            stabilizers would be redundant.

        Warning:
            This function uses the :class:`~tqec.plaquette.enums.PlaquetteOrientation`
            class. For a 2-body stabilizer measurement plaquette, the "orientation"
            corresponds to the direction in which the rounded side is pointing.
            So a plaquette with the orientation ``PlaquetteOrientation.DOWN`` has the
            following properties:
            - it measures the 2 data-qubits on the **top** side of the usual 4-body
            stabilizer measurement plaquette,
            - it can be used for a bottom boundary,
            - its rounded side points downwards.

        Returns:
            a mapping with 8 plaquettes: one for each basis (either ``X`` or ``Z``)
            and for each plaquette orientation (``UP``, ``DOWN``, ``LEFT`` or
            ``RIGHT``).

        """
        _po = PlaquetteOrientation
        h = "h" if hadamard else "-"
        ret: dict[Basis, dict[PlaquetteOrientation, RPNGDescription]] = {}
        # Note: the schedule of CNOT gates in weight-2 plaquettes is less
        # important because hook errors do not exist. We arbitrarily chose the
        # vertical schedule.
        s = VERTICAL_HOOK_SCHEDULES[is_reversed]
        for basis in Basis:
            b = basis.value.lower()
            ret[basis] = {
                _po.DOWN: RPNGDescription.from_string(f"-{b}{s[0]}{h} -{b}{s[1]}{h} ---- ----"),
                _po.LEFT: RPNGDescription.from_string(f"---- -{b}{s[1]}{h} ---- -{b}{s[3]}{h}"),
                _po.UP: RPNGDescription.from_string(f"---- ---- -{b}{s[2]}{h} -{b}{s[3]}{h}"),
                _po.RIGHT: RPNGDescription.from_string(f"-{b}{s[0]}{h} ---- -{b}{s[2]}{h} ----"),
            }
        return ret

    def get_extended_plaquettes(
        self, reset: Basis | None, measurement: Basis | None, is_reversed: bool
    ) -> dict[Basis, ExtendedPlaquetteCollection]:
        """Get plaquettes that are supposed to be used to implement ``UP`` or ``DOWN`` spatial
        pipes.

        Returns:
            a map from stabilizer basis to :class:`ExtendedPlaquetteCollection`.

        """
        return {
            b: (ExtendedPlaquetteCollection.from_basis(b, reset, measurement, is_reversed))
            for b in Basis
        }

    def get_bulk_hadamard_rpng_descriptions(
        self, is_reversed: bool
    ) -> dict[Basis, dict[Orientation, RPNGDescription]]:
        """Get plaquettes that are supposed to be used to implement a temporal Hadamard pipe.

        Returns:
            a mapping with 4 plaquettes: one for each basis (either ``X`` or ``Z``)
            and for each hook orientation (either ``HORIZONTAL`` or ``VERTICAL``).

        """
        # 2-qubit gate schedules
        vsched = VERTICAL_HOOK_SCHEDULES[is_reversed]
        hsched = HORIZONTAL_HOOK_SCHEDULES[is_reversed]
        return {
            basis: {
                orientation: RPNGDescription.from_basis_and_schedule(
                    basis, sched, measurement=ExtendedBasis.H
                )
                for orientation, sched in (
                    (Orientation.VERTICAL, vsched),
                    (Orientation.HORIZONTAL, hsched),
                )
            }
            for basis in Basis
        }

    def get_spatial_x_hadamard_rpng_descriptions(
        self,
        top_left_basis: Basis,
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> tuple[RPNGDescription, RPNGDescription, RPNGDescription]:
        """Return a description of the 3 different plaquettes needed to perform a spatial hadamard
        transformation between two qubits aligned on the X axis.

        Args:
            top_left_basis: basis of the stabilizer measured by the top-left
                data-qubit.
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            reset: basis of the reset operation performed on data-qubits. Defaults
                to ``None`` that translates to no reset being applied on data-qubits.
            measurement: basis of the measurement operation performed on data-qubits.
                Defaults to ``None`` that translates to no measurement being applied
                on data-qubits.

        Returns:
            a tuple ``(bulk1, bulk2, bottom)`` containing:

            - ``bulk1``, a square plaquette with its two left-most data-qubits
              measuring ``top_left_basis`` stabilizer.
            - ``bulk2``, a square plaquette with its two left-most data-qubits
              measuring ``top_left_basis.flipped()`` stabilizer.
            - ``bottom``, a plaquette measuring a weight 2 stabilizer with its
              left-most data-qubit measuring ``top_left_basis`` stabilizer and
              right-most data-qubit measuring ``top_left_basis.flipped()``
              stabilizer.

        """
        b = top_left_basis.value.lower()
        o = top_left_basis.flipped().value.lower()
        r = reset.value.lower() if reset is not None else "-"
        m = measurement.value.lower() if measurement is not None else "-"
        # 2-qubit gate schedules
        vs = VERTICAL_HOOK_SCHEDULES[is_reversed]
        hs = HORIZONTAL_HOOK_SCHEDULES[is_reversed]
        return (
            RPNGDescription.from_string(
                f"-{b}{hs[0]}- {r}{o}{hs[1]}{m} -{b}{hs[2]}- {r}{o}{hs[3]}{m}"
            ),
            RPNGDescription.from_string(
                f"-{o}{vs[0]}- {r}{b}{vs[1]}{m} -{o}{vs[2]}- {r}{b}{vs[3]}{m}"
            ),
            RPNGDescription.from_string(f"-{b}{hs[0]}- {r}{o}{hs[1]}{m} ---- ----"),
        )

    def get_spatial_y_hadamard_rpng_descriptions(
        self,
        top_left_basis: Basis,
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> tuple[RPNGDescription, RPNGDescription, RPNGDescription]:
        """Return a description of the 3 different plaquettes needed to perform a spatial hadamard
        transformation between two qubits aligned on the Y axis.

        Args:
            top_left_basis: basis of the top-left-most stabilizer (top
                stabilizer of a 2-qubit plaquette).
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            reset: basis of the reset operation performed on data-qubits. Defaults
                to ``None`` that translates to no reset being applied on data-qubits.
            measurement: basis of the measurement operation performed on data-qubits.
                Defaults to ``None`` that translates to no measurement being applied
                on data-qubits.

        Returns:
            a tuple ``(bulk1, bulk2, left)`` containing:

            - ``bulk1``, a square plaquette with its two top-most data-qubits
              measuring ``top_left_basis.flipped()`` stabilizer.
            - ``bulk2``, a square plaquette with its two top-most data-qubits
              measuring ``top_left_basis`` stabilizer.
            - ``left``, a plaquette measuring a weight 2 stabilizer with its
              top-most data-qubit measuring ``top_left_basis`` stabilizer and
              bottom-most data-qubit measuring ``top_left_basis.flipped()``
              stabilizer.

        Warning:
            When seen visually, the plaquettes returned by this method are not
            in reading order. Visually, the order is::

                left  |  bulk1  |  bulk2

            but this method returns ``(bulk1, bulk2, left)``.

        """
        b = top_left_basis.value.lower()
        o = top_left_basis.flipped().value.lower()
        r = reset.value.lower() if reset is not None else "-"
        m = measurement.value.lower() if measurement is not None else "-"
        # 2-qubit gate schedules
        vs = VERTICAL_HOOK_SCHEDULES[is_reversed]
        hs = HORIZONTAL_HOOK_SCHEDULES[is_reversed]
        return (
            RPNGDescription.from_string(
                f"-{o}{hs[0]}- -{o}{hs[1]}- {r}{b}{hs[2]}{m} {r}{b}{hs[3]}{m}"
            ),
            RPNGDescription.from_string(
                f"-{b}{vs[0]}- -{b}{vs[1]}- {r}{o}{vs[2]}{m} {r}{o}{vs[3]}{m}"
            ),
            RPNGDescription.from_string(f"---- -{b}{vs[1]}- ---- {r}{o}{vs[3]}{m}"),
        )

    ############################################################
    #                          Memory                          #
    ############################################################

    ########################################
    #             Regular qubit            #
    ########################################
    def get_memory_qubit_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.RectangularTemplate` instance needed to
        implement a single logical qubit.
        """
        return QubitTemplate()

    def get_memory_qubit_rpng_descriptions(
        self,
        is_reversed: bool,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a standard memory operation
        on a logical qubit.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_memory_qubit_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned logical qubit description.
            reset: basis of the reset operation performed on data-qubits.
                Defaults to ``None`` that translates to no reset being applied
                on data-qubits.
            measurement: basis of the measurement operation performed on
                data-qubits. Defaults to ``None`` that translates to no
                measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a standard
            memory operation on a logical qubit, optionally with resets or
            measurements on the data-qubits too.

        """
        # Basis for top/bottom and left/right boundary plaquettes
        hbasis = Basis.Z if z_orientation == Orientation.HORIZONTAL else Basis.X
        vbasis = hbasis.flipped()
        # Generating plaquette descriptions we will need later
        bulk_descriptions = self.get_bulk_rpng_descriptions(is_reversed, reset, measurement)
        two_body_descriptions = self.get_2_body_rpng_descriptions(is_reversed)
        return FrozenDefaultDict(
            {
                6: two_body_descriptions[vbasis][PlaquetteOrientation.UP],
                7: two_body_descriptions[hbasis][PlaquetteOrientation.LEFT],
                # Bulk
                9: bulk_descriptions[vbasis][Orientation.HORIZONTAL],
                10: bulk_descriptions[hbasis][Orientation.VERTICAL],
                12: two_body_descriptions[hbasis][PlaquetteOrientation.RIGHT],
                13: two_body_descriptions[vbasis][PlaquetteOrientation.DOWN],
            },
            default_value=RPNGDescription.empty(),
        )

    def get_memory_qubit_plaquettes(
        self,
        is_reversed: bool,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a standard memory operation on a logical
        qubit.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_memory_qubit_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned logical qubit description.
            reset: basis of the reset operation performed on data-qubits.
                Defaults to ``None`` that translates to no reset being applied
                on data-qubits.
            measurement: basis of the measurement operation performed on
                data-qubits. Defaults to ``None`` that translates to no
                measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a standard
            memory operation on a logical qubit, optionally with resets or
            measurements on the data-qubits too.

        """
        return self._mapper(self.get_memory_qubit_rpng_descriptions)(
            is_reversed, z_orientation, reset, measurement
        )

    ########################################
    #                X pipe                #
    ########################################
    def get_memory_vertical_boundary_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.RectangularTemplate` instance needed to
        implement a regular spatial pipe between two logical qubits aligned on the ``X`` axis.
        """
        return QubitVerticalBorders()

    def get_memory_vertical_boundary_rpng_descriptions(
        self,
        is_reversed: bool,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a standard memory operation
        on a pipe between two neighbouring logical qubits aligned on the ``X``-axis.

        Note:
            if ``reset`` (resp. ``measurement``) is not ``None``, a reset (resp.
            measurement) operation in the provided basis will be inserted **only
            on internal data-qubits**. Here, internal data-qubits are all the
            qubits that are in the middle of the template.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_memory_vertical_boundary_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned memory description.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a standard memory
            operation on a pipe between two neighbouring logical qubits aligned on
            the ``X``-axis, optionally with resets or measurements on the
            data-qubits too.

        """
        # Basis for top/bottom boundary plaquettes
        vbasis = Basis.Z if z_orientation == Orientation.VERTICAL else Basis.X
        hbasis = vbasis.flipped()
        # Generating plaquette descriptions we will need
        left_bulk_descriptions = self.get_bulk_rpng_descriptions(
            is_reversed, reset, measurement, (1, 3)
        )
        right_bulk_descriptions = self.get_bulk_rpng_descriptions(
            is_reversed, reset, measurement, (0, 2)
        )
        two_body_descriptions = self.get_2_body_rpng_descriptions(is_reversed)

        return FrozenDefaultDict(
            {
                2: two_body_descriptions[vbasis][PlaquetteOrientation.UP],
                3: two_body_descriptions[vbasis][PlaquetteOrientation.DOWN],
                # LEFT bulk
                5: left_bulk_descriptions[vbasis][Orientation.HORIZONTAL],
                6: left_bulk_descriptions[hbasis][Orientation.VERTICAL],
                # RIGHT bulk
                7: right_bulk_descriptions[hbasis][Orientation.VERTICAL],
                8: right_bulk_descriptions[vbasis][Orientation.HORIZONTAL],
            },
            default_value=RPNGDescription.empty(),
        )

    def get_memory_vertical_boundary_plaquettes(
        self,
        is_reversed: bool,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a standard memory operation on a pipe between
        two neighbouring logical qubits aligned on the ``X``-axis.

        Note:
            if ``reset`` (resp. ``measurement``) is not ``None``, a reset (resp.
            measurement) operation in the provided basis will be inserted **only
            on internal data-qubits**. Here, internal data-qubits are all the
            qubits that are in the middle of the template.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_memory_vertical_boundary_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned memory description.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            the plaquettes needed to implement a standard memory operation on a
            pipe between two neighbouring logical qubits aligned on the
            ``X``-axis, optionally with resets or measurements on the
            data-qubits too.

        """
        return self._mapper(self.get_memory_vertical_boundary_rpng_descriptions)(
            is_reversed, z_orientation, reset, measurement
        )

    ########################################
    #                Y pipe                #
    ########################################
    def get_memory_horizontal_boundary_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.RectangularTemplate` instance needed to
        implement a regular spatial pipe between two logical qubits aligned on the ``Y`` axis.
        """
        return QubitHorizontalBorders()

    def get_memory_horizontal_boundary_rpng_descriptions(
        self,
        is_reversed: bool,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a standard memory operation
        on a pipe between two neighbouring logical qubits aligned on the ``Y``-axis.

        Note:
            if ``reset`` (resp. ``measurement``) is not ``None``, a reset (resp.
            measurement) operation in the provided basis will be inserted **only
            on internal data-qubits**. Here, internal data-qubits are all the
            qubits that are in the middle of the template.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_memory_horizontal_boundary_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned memory description.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a standard memory
            operation on a pipe between two neighbouring logical qubits aligned on
            the ``Y``-axis, optionally with resets or measurements on the
            data-qubits too.

        """
        # Basis for left/right boundary plaquettes
        hbasis = Basis.Z if z_orientation == Orientation.HORIZONTAL else Basis.X
        vbasis = hbasis.flipped()
        # Generating plaquette descriptions we will need later
        up_bulk_descriptions = self.get_bulk_rpng_descriptions(
            is_reversed, reset, measurement, (2, 3)
        )
        down_bulk_descriptions = self.get_bulk_rpng_descriptions(
            is_reversed, reset, measurement, (0, 1)
        )
        two_body_descriptions = self.get_2_body_rpng_descriptions(is_reversed)

        return FrozenDefaultDict(
            {
                1: two_body_descriptions[hbasis][PlaquetteOrientation.LEFT],
                4: two_body_descriptions[hbasis][PlaquetteOrientation.RIGHT],
                # TOP bulk
                5: up_bulk_descriptions[vbasis][Orientation.HORIZONTAL],
                6: up_bulk_descriptions[hbasis][Orientation.VERTICAL],
                # BOTTOM bulk
                7: down_bulk_descriptions[hbasis][Orientation.VERTICAL],
                8: down_bulk_descriptions[vbasis][Orientation.HORIZONTAL],
            },
            default_value=RPNGDescription.empty(),
        )

    def get_memory_horizontal_boundary_plaquettes(
        self,
        is_reversed: bool,
        z_orientation: Orientation = Orientation.HORIZONTAL,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a standard memory operation on a pipe between
        two neighbouring logical qubits aligned on the ``Y``-axis.

        Note:
            if ``reset`` (resp. ``measurement``) is not ``None``, a reset (resp.
            measurement) operation in the provided basis will be inserted **only
            on internal data-qubits**. Here, internal data-qubits are all the
            qubits that are in the middle of the template.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_memory_horizontal_boundary_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            z_orientation: orientation of the ``Z`` observable. Used to compute
                the stabilizers that should be measured on the boundaries and in
                the bulk of the returned memory description.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            the plaquettes needed to implement a standard memory operation on a
            pipe between two neighbouring logical qubits aligned on the
            ``Y``-axis, optionally with resets or measurements on the
            data-qubits too.

        """
        return self._mapper(self.get_memory_horizontal_boundary_rpng_descriptions)(
            is_reversed, z_orientation, reset, measurement
        )

    ############################################################
    #                          Spatial                         #
    ############################################################
    def get_spatial_cube_qubit_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.RectangularTemplate` instance needed to
        implement a spatial cube.

        Note:
            A spatial cube is defined as a cube with all its spatial boundaries
            in the same basis.
            Such a cube might appear in stability experiments (e.g.,
            http://arxiv.org/abs/2204.13834), in spatial junctions (i.e., a cube
            with more than one pipe in the spatial plane) or in other QEC gadgets
            such as the lattice surgery implementation of a ``CZ`` gate.

        """
        return QubitSpatialCubeTemplate()

    def get_spatial_cube_qubit_rpng_descriptions(
        self,
        spatial_boundary_basis: Basis,
        arms: SpatialArms,
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a spatial cube.

        Note:
            A spatial cube is defined as a cube with all its spatial boundaries
            in the same basis.
            Such a cube might appear in stability experiments (e.g.,
            http://arxiv.org/abs/2204.13834), in spatial junctions (i.e., a cube
            with more than one pipe in the spatial plane) or in other QEC
            gadgets such as the lattice surgery implementation of a ``CZ`` gate.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_spatial_cube_qubit_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            spatial_boundary_basis: stabilizers that are measured at each
                boundaries of the spatial cube.
            arms: flag-like enumeration listing the arms that are used around
                the logical qubit. The returned template will be adapted to be
                compatible with such a layout.
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            reset: basis of the reset operation performed on data-qubits.
                Defaults to ``None`` that translates to no reset being applied
                on data-qubits.
            measurement: basis of the measurement operation performed on
                data-qubits. Defaults to ``None`` that translates to no
                measurement being applied on data-qubits.

        Raises:
            TQECError: if ``arms`` describes an I-shaped junction (TOP/DOWN
                or LEFT/RIGHT).

        Returns:
            a description of the plaquettes needed to implement a spatial cube.

        """
        # In this function implementation, all the indices used are referring to the
        # indices returned by the QubitSpatialCubeTemplate template. They are
        # copied below for convenience, but the only source of truth is in the
        # QubitSpatialCubeTemplate docstring!
        #      1   9  10   9  10   9  10   9  10   2
        #     11   5  17  13  17  13  17  13   6  21
        #     12  20  13  17  13  17  13  17  14  22
        #     11  16  20  13  17  13  17  14  18  21
        #     12  20  16  20  13  17  14  18  14  22
        #     11  16  20  16  19  15  18  14  18  21
        #     12  20  16  19  15  19  15  18  14  22
        #     11  16  19  15  19  15  19  15  18  21
        #     12   7  15  19  15  19  15  19   8  22
        #      3  23  24  23  24  23  24  23  24   4
        if arms in SpatialArms.I_shaped_arms():
            raise TQECError(
                "I-shaped spatial junctions (i.e., spatial junctions with only two "
                "arms that are the opposite of each other: LEFT/RIGHT or UP/DOWN) "
                "should not use get_spatial_cube_qubit_template but rather use "
                "a conventional memory logical qubit with get_memory_qubit_template."
            )
        # _sbb: Spatial Boundary Basis.
        _sbb = spatial_boundary_basis
        # Pre-define some collection of plaquette descriptions
        corner_descriptions = self.get_3_body_rpng_descriptions(
            _sbb, is_reversed, reset, measurement
        )
        bulk_descriptions = self.get_bulk_rpng_descriptions(is_reversed, reset, measurement)
        two_body_descriptions = self.get_2_body_rpng_descriptions(is_reversed)

        if arms == SpatialArms.NONE:
            # Stability experiment
            return FrozenDefaultDict(
                {
                    5: corner_descriptions[0],
                    6: bulk_descriptions[_sbb.flipped()][Orientation.VERTICAL],
                    7: bulk_descriptions[_sbb.flipped()][Orientation.VERTICAL],
                    8: corner_descriptions[3],
                    10: two_body_descriptions[_sbb][PlaquetteOrientation.UP],
                    12: two_body_descriptions[_sbb][PlaquetteOrientation.LEFT],
                    13: bulk_descriptions[_sbb][Orientation.HORIZONTAL],
                    14: bulk_descriptions[_sbb][Orientation.VERTICAL],
                    15: bulk_descriptions[_sbb][Orientation.HORIZONTAL],
                    16: bulk_descriptions[_sbb][Orientation.VERTICAL],
                    17: bulk_descriptions[_sbb.flipped()][Orientation.VERTICAL],
                    18: bulk_descriptions[_sbb.flipped()][Orientation.HORIZONTAL],
                    19: bulk_descriptions[_sbb.flipped()][Orientation.VERTICAL],
                    20: bulk_descriptions[_sbb.flipped()][Orientation.HORIZONTAL],
                    21: two_body_descriptions[_sbb][PlaquetteOrientation.RIGHT],
                    23: two_body_descriptions[_sbb][PlaquetteOrientation.DOWN],
                },
                default_value=RPNGDescription.empty(),
            )
        # Note about the fixed boundary convention: in order to work as expected,
        # spatial cubes need to have one dimension that does not respect the
        # boundary convention. By convention, we only use stretched stabilizers in
        # the vertical (Y) dimension (i.e., between two cubes that are aligned
        # on the Y axis), and so only the boundaries on the X axis (left and
        # right) need to not follow the fixed boundary convention.
        # For spatial cubes, the only exception to the above rule is when the
        # cube is a "dead-end" (i.e., only one spatial arm: ``len(arms) == 1``).
        # In that case, the dimension that should not follow the fixed boundary
        # convention is the one "closing" the pipe, i.e., the dimension in which
        # the only arm is positioned.
        # For dead-end cubes, a dead-end in the X dimension still follows the
        # general rule (the boundaries in the X axis do not follow the
        # convention), so we only have to test if we have a dead-end in the Y
        # dimension.
        odd_boundary_dimension: Final[Literal[Direction3D.X, Direction3D.Y]] = (
            Direction3D.Y if arms in [SpatialArms.UP, SpatialArms.DOWN] else Direction3D.X
        )

        mapping: dict[int, RPNGDescription] = {}

        ####################
        #    Boundaries    #
        ####################
        # Fill the boundaries that should be filled in the returned template
        # because they have no arms, and so will not be filled later.
        top, bottom, left, right = (
            (10, 23, 12, 21) if odd_boundary_dimension == Direction3D.X else (9, 24, 11, 22)
        )
        if SpatialArms.UP not in arms:
            mapping[top] = two_body_descriptions[_sbb][PlaquetteOrientation.UP]
        if SpatialArms.RIGHT not in arms:
            mapping[right] = two_body_descriptions[_sbb][PlaquetteOrientation.RIGHT]
        if SpatialArms.DOWN not in arms:
            mapping[bottom] = two_body_descriptions[_sbb][PlaquetteOrientation.DOWN]
        if SpatialArms.LEFT not in arms:
            mapping[left] = two_body_descriptions[_sbb][PlaquetteOrientation.LEFT]

        ####################
        #       Bulk       #
        ####################
        # Bulk plaquettes basis might change according to the odd boundary
        # dimension to avoid having 2 plaquettes measuring the same basis side
        # by side.
        # tlb, otb: Top-Left (plaquette) Basis, Other Basis (for the bulk)
        tlb = _sbb if odd_boundary_dimension == Direction3D.X else _sbb.flipped()
        otb = tlb.flipped()
        # Assigning plaquette description to the bulk, considering that the bulk
        # corners (i.e. indices {5, 6, 7, 8}) should be assigned "regular" plaquettes
        # (i.e. 6 is assigned the same plaquette as 17, 7 -> 19, 5 -> 13, 8 -> 15).
        # If these need to be changed, it will be done afterwards.
        # Setting the orientations for SBB plaquettes for each of the four
        # portions of the template bulk.
        sbb_up = sbb_down = Orientation.VERTICAL
        sbb_right = sbb_left = Orientation.HORIZONTAL
        # If the corresponding arm is missing, the SBB plaquette hook error
        # orientation should flip to avoid shortcuts due to hook errors.
        sbb_up = sbb_up if SpatialArms.UP in arms else sbb_up.flip()
        sbb_down = sbb_down if SpatialArms.DOWN in arms else sbb_down.flip()
        sbb_right = sbb_right if SpatialArms.RIGHT in arms else sbb_right.flip()
        sbb_left = sbb_left if SpatialArms.LEFT in arms else sbb_left.flip()
        # The OTH (other basis) orientations are the opposite of the SBB
        # orientation.
        oth_up, oth_down = sbb_up.flip(), sbb_down.flip()
        oth_right, oth_left = sbb_right.flip(), sbb_left.flip()

        # Setting the SBB plaquettes
        mapping[5] = mapping[13] = bulk_descriptions[tlb][sbb_up]
        mapping[8] = mapping[15] = bulk_descriptions[tlb][sbb_down]
        mapping[14] = bulk_descriptions[tlb][sbb_right]
        mapping[16] = bulk_descriptions[tlb][sbb_left]
        # Setting the OTH plaquettes
        mapping[6] = mapping[17] = bulk_descriptions[otb][oth_up]
        mapping[7] = mapping[19] = bulk_descriptions[otb][oth_down]
        mapping[18] = bulk_descriptions[otb][oth_right]
        mapping[20] = bulk_descriptions[otb][oth_left]

        # For the in-bulk corners, if the two arms around the corner are not
        # present, the corner plaquette has been removed from the mapping. The
        # corner **within the bulk** should be overwritten to become a 3-body
        # stabilizer measurement.
        if arms == SpatialArms.RIGHT:
            mapping[5] = corner_descriptions[0]
        elif arms == SpatialArms.DOWN:
            mapping[6] = corner_descriptions[1]
        elif arms == SpatialArms.UP:
            mapping[7] = corner_descriptions[2]
        elif arms == SpatialArms.LEFT:
            mapping[8] = corner_descriptions[3]
        # At this point, we are sure that len(arms) >= 2. The only cases left
        # where a 3-body stabilizer is needed are the following:
        if arms == SpatialArms.RIGHT | SpatialArms.DOWN:
            mapping[5] = corner_descriptions[0]
        if arms == SpatialArms.LEFT | SpatialArms.UP:
            mapping[8] = corner_descriptions[3]

        ####################
        #  Sanity checks   #
        ####################
        # All the plaquettes in the bulk should be set.
        bulk_plaquette_indices = set(range(5, 9)) | set(range(13, 21))
        missing_bulk_plaquette_indices = bulk_plaquette_indices - mapping.keys()
        assert not missing_bulk_plaquette_indices, (
            "Some plaquette(s) in the bulk were not correctly assigned to a "
            f"RPNGDescription. Missing indices: {missing_bulk_plaquette_indices}."
        )
        return FrozenDefaultDict(mapping, default_value=RPNGDescription.empty())

    def get_spatial_cube_qubit_plaquettes(
        self,
        spatial_boundary_basis: Basis,
        arms: SpatialArms,
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a spatial cube.

        Note:
            A spatial cube is defined as a cube with all its spatial boundaries
            in the same basis.
            Such a cube might appear in stability experiments (e.g.,
            http://arxiv.org/abs/2204.13834), in spatial junctions (i.e., a cube
            with more than one pipe in the spatial plane) or in other QEC
            gadgets such as the lattice surgery implementation of a ``CZ`` gate.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_spatial_cube_qubit_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            spatial_boundary_basis: stabilizers that are measured at each
                boundaries of the spatial cube.
            arms: flag-like enumeration listing the arms that are used around
                the logical qubit. The returned template will be adapted to be
                compatible with such a layout.
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            reset: basis of the reset operation performed on data-qubits.
                Defaults to ``None`` that translates to no reset being applied
                on data-qubits.
            measurement: basis of the measurement operation performed on
                data-qubits. Defaults to ``None`` that translates to no
                measurement being applied on data-qubits.

        Raises:
            TQECError: if ``arms`` only contains 0 or 1 flag.
            TQECError: if ``arms`` describes an I-shaped junction (TOP/DOWN
                or LEFT/RIGHT).

        Returns:
            the plaquettes needed to implement a spatial cube.

        """
        return self._mapper(self.get_spatial_cube_qubit_rpng_descriptions)(
            spatial_boundary_basis, arms, is_reversed, reset, measurement
        )

    ########################################
    #              Spatial arm             #
    ########################################
    def get_spatial_cube_arm_raw_template(self, arms: SpatialArms) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.RectangularTemplate` instance needed to
        implement the given spatial ``arms``.

        Args:
            arms: specification of the spatial arm(s) we want a template for.
                Needs to contain either one arm, or 2 arms that form a line
                (e.g., ``SpatialArms.UP | SpatialArms.DOWN``).
                If the arm that should be built has one spatial cube on top,
                this should be ``SpatialArm.DOWN`` because it is the bottom arm
                of a spatial cube.
                If the arm links 2 spatial cubes, the ``arms`` parameter should
                be the union of the arms (and so can only be
                ``SpatialArms.UP | SpatialArms.DOWN`` or
                ``SpatialArms.LEFT | SpatialArms.RIGHT`` because any other
                combination cannot be formed by a single arm).

        Raises:
            TQECError: if the provided ``arms`` value does not check the
                documented pre-conditions.

        """
        if (
            len(arms) == 0
            or len(arms) > 2
            or (len(arms) == 2 and arms not in SpatialArms.I_shaped_arms())
        ):
            raise TQECError(
                f"The two provided arms cannot form a spatial pipe. Got {arms} but "
                f"expected either a single {SpatialArms.__name__} or two but in a "
                f"line (e.g., {SpatialArms.I_shaped_arms()})."
            )
        if SpatialArms.LEFT in arms or SpatialArms.RIGHT in arms:
            return QubitVerticalBorders()
        elif SpatialArms.UP is arms or SpatialArms.DOWN in arms:
            return QubitHorizontalBorders()
        else:
            raise TQECError(f"Unrecognized spatial arm(s): {arms}.")

    def get_spatial_cube_arm_plaquettes(
        self,
        spatial_boundary_basis: Basis,
        arms: SpatialArms,
        linked_cubes: tuple[CubeSpec, CubeSpec],
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement **one** pipe connecting to a spatial cube.

        Note:
            A spatial cube is defined as a cube with all its spatial boundaries
            in the same basis.
            Such a cube might appear in stability experiments (e.g.,
            http://arxiv.org/abs/2204.13834), in spatial junctions (i.e., a cube
            with more than one pipe in the spatial plane) or in other QEC gadgets
            such as the lattice surgery implementation of a ``CZ`` gate.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_spatial_cube_arm_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            spatial_boundary_basis: stabilizers that are measured at each
                boundaries of the spatial cube.
            arms: arm(s) of the spatial cube(s) linked by the pipe.
            linked_cubes: a tuple ``(u, v)`` where ``u`` and ``v`` are the
                specifications of the two ends of the pipe to generate RPNG
                descriptions for.
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Raises:
            TQECError: if ``arm`` does not contain exactly 1 or 2 flags (i.e.,
                if it contains 0 or 3+ flags).

        Returns:
            the plaquettes needed to implement **one** pipe connecting to a
            spatial cube.

        """
        if len(arms) == 2 and arms not in SpatialArms.I_shaped_arms():
            raise TQECError(
                f"The two provided arms cannot form a spatial pipe. Got {arms} but "
                f"expected either a single {SpatialArms.__name__} or two but in a "
                f"line (e.g., {SpatialArms.I_shaped_arms()})."
            )
        if arms in [
            SpatialArms.LEFT,
            SpatialArms.RIGHT,
            SpatialArms.LEFT | SpatialArms.RIGHT,
        ]:
            return self._get_left_right_spatial_cube_arm_plaquettes(
                spatial_boundary_basis,
                arms,
                linked_cubes,
                is_reversed,
                reset,
                measurement,
            )
        if arms in [
            SpatialArms.UP,
            SpatialArms.DOWN,
            SpatialArms.UP | SpatialArms.DOWN,
        ]:
            return self._get_up_down_spatial_cube_arm_plaquettes(
                spatial_boundary_basis,
                linked_cubes,
                is_reversed,
                reset,
                measurement,
            )
        raise TQECError(f"Got an invalid arm: {arms}.")

    def _get_left_right_spatial_cube_arm_plaquettes(
        self,
        spatial_boundary_basis: Basis,
        arms: SpatialArms,
        linked_cubes: tuple[CubeSpec, CubeSpec],
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        # This is a regular memory arm, except that we should make sure that one
        # of the boundary does not override the extended stabilizer.
        z_orientation = (
            Orientation.VERTICAL if spatial_boundary_basis == Basis.Z else Orientation.HORIZONTAL
        )
        regular_memory = self.get_memory_vertical_boundary_plaquettes(
            is_reversed, z_orientation, reset, measurement
        )
        u, v = linked_cubes
        if SpatialArms.LEFT in arms and SpatialArms.UP in v.spatial_arms:
            regular_memory = regular_memory.without_plaquettes([2])
        if SpatialArms.RIGHT in arms and SpatialArms.DOWN in u.spatial_arms:
            regular_memory = regular_memory.without_plaquettes([3])
        return regular_memory

    @staticmethod
    def pipe_needs_extended_stabilizers(linked_cubes: tuple[CubeSpec, CubeSpec]) -> bool:
        """Check if the pipe represented by the given ``arms`` and ``linked_cubes`` requires
        extended stabilizers.

        In fixed boundary convention, spatial cubes change the boundary parity.
        That is why we need stretched stabilizers. By convention, TQEC inserts
        stretched stabilizers only in the UP/DOWN pipes (i.e., in the Y spatial dimension).

        But if 2 spatial cubes are linked by a pipe in the Y dimension, we
        *might* not need to fix the parity with extended stabilizers. We only
        need to use stretched stabilizers when the parity would be wrong, and
        that is only when exactly 1 of the 2 cubes linked by the pipe has pipes
        in both spatial dimensions

        Args:
            linked_cubes: a tuple ``(u, v)`` where ``u`` and ``v`` are the
                specifications of the two ends of the pipe.

        Returns:
            ``True`` if extended stabilizers should be used, ``False`` otherwise.

        """
        return (
            linked_cubes[0].has_spatial_pipe_in_both_dimensions
            ^ linked_cubes[1].has_spatial_pipe_in_both_dimensions
        )

    @staticmethod
    def pipe_has_boundary_extended_stabilizer_at_left(
        linked_cubes: tuple[CubeSpec, CubeSpec],
    ) -> bool:
        """Check if the pipe represented by the given ``linked_cubes`` has an extended stabilizer at
        the left boundary.

        Assuming a pipe needs to be implemented with extended stabilizer, it has
        an extended stabilizer at the left boundary iff. the bottom cube of the
        pipe has arms in both spatial directions and the top cube does not.

        Args:
            linked_cubes: a tuple ``(u, v)`` where ``u`` and ``v`` are the
                specifications of the two ends of the pipe.

        Returns:
            ``True`` if an extended stabilizer should be used at the left
            boundary, ``False`` otherwise.

        """
        assert FixedBoundaryConventionGenerator.pipe_needs_extended_stabilizers(linked_cubes)
        return linked_cubes[1].has_spatial_pipe_in_both_dimensions

    def _get_up_down_spatial_cube_arm_plaquettes(
        self,
        spatial_boundary_basis: Basis,
        linked_cubes: tuple[CubeSpec, CubeSpec],
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        if not FixedBoundaryConventionGenerator.pipe_needs_extended_stabilizers(linked_cubes):
            # Special case, a little bit simpler, not using extended stabilizers.
            return self._get_up_and_down_spatial_cube_arm_plaquettes(
                spatial_boundary_basis, linked_cubes, is_reversed, reset, measurement
            )
        # General case, need extended stabilizers.
        sbb, otb = spatial_boundary_basis, spatial_boundary_basis.flipped()
        # EPs: extended plaquettes
        extended_plaquettes = self.get_extended_plaquettes(reset, measurement, is_reversed)
        # Dictionary that will be filled with plaquettes
        plaquettes: dict[int, Plaquette] = {}
        # Getting the extended plaquettes for the bulk and filling the dictionary
        has_left_boundary = (
            FixedBoundaryConventionGenerator.pipe_has_boundary_extended_stabilizer_at_left(
                linked_cubes
            )
        )
        bulk1 = extended_plaquettes[otb if has_left_boundary else sbb].bulk
        bulk2 = extended_plaquettes[sbb if has_left_boundary else otb].bulk
        plaquettes |= {5: bulk1.top, 6: bulk2.top, 7: bulk1.bottom, 8: bulk2.bottom}
        # Getting the extended plaquette, either for the left or the right
        # boundary depending on the spatial arm that is being asked for.
        boundary_collection = extended_plaquettes[sbb]
        u, v = linked_cubes
        if has_left_boundary:
            boundary = (
                boundary_collection.left_with_arm
                if SpatialArms.LEFT in v.spatial_arms
                else boundary_collection.left_without_arm
            )
            plaquettes |= {1: boundary.top, 3: boundary.bottom}
        else:
            boundary = (
                boundary_collection.right_with_arm
                if SpatialArms.RIGHT in u.spatial_arms
                else boundary_collection.right_without_arm
            )
            plaquettes |= {2: boundary.top, 4: boundary.bottom}
        return Plaquettes(
            FrozenDefaultDict(
                plaquettes,
                default_value=self._mapper.get_plaquette(RPNGDescription.empty()),
            )
        )

    def _get_up_and_down_spatial_cube_arm_plaquettes(
        self,
        spatial_boundary_basis: Basis,
        linked_cubes: tuple[CubeSpec, CubeSpec],
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        return self._mapper(self._get_up_and_down_spatial_cube_arm_rpng_descriptions)(
            spatial_boundary_basis, linked_cubes, is_reversed, reset, measurement
        )

    def _get_up_and_down_spatial_cube_arm_rpng_descriptions(
        self,
        spatial_boundary_basis: Basis,
        linked_cubes: tuple[CubeSpec, CubeSpec],
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return the RPNG descriptions to implement a pipe connecting at least one spatial cube.

        The pipe implemented by this method links two cubes such as:

        - at least one of the two cube is a spatial cube (both can be),
        - either none of both of the two linked cubes have pipes in both spatial
          dimensions.

        In particular, the following situations can be encountered (list is not
        exhaustive):

        - a straight line ending on a spatial cube, meaning that the pipe links
          a spatial cube with a single arm and a regular cube (case where none
          of the 2 linked cubes have pipes in both spatial dimensions),
        - a "rotated-H" shape (case where both of the 2 linked cubes have pipes
          in both spatial dimensions),
        - ...

        These pipes have in common the fact that they do not require extended
        stabilizers to be implemented.

        """
        # Aliases to shorten line length.
        r, m = reset, measurement
        _sbb = spatial_boundary_basis
        # Generating the plaquette descriptions we will need later
        up_bulk_plaquettes = self.get_bulk_rpng_descriptions(is_reversed, r, m, (2, 3))
        down_bulk_plaquettes = self.get_bulk_rpng_descriptions(is_reversed, r, m, (0, 1))
        corner_descriptions = self.get_3_body_rpng_descriptions(_sbb, is_reversed, r, m)
        two_body_descriptions = self.get_2_body_rpng_descriptions(is_reversed)
        # Here, depending on the linked cubes, we might insert regular two-body
        # plaquettes or three-body plaquettes.
        u, v = linked_cubes
        both_cubes_have_spatial_pipes_in_both_dimensions = (
            u.has_spatial_pipe_in_both_dimensions and v.has_spatial_pipe_in_both_dimensions
        )
        right_plaquette = (
            corner_descriptions[3]
            if SpatialArms.RIGHT in u.spatial_arms
            else two_body_descriptions[_sbb][PlaquetteOrientation.RIGHT]
        )
        left_plaquette = (
            corner_descriptions[0]
            if SpatialArms.LEFT in v.spatial_arms
            else two_body_descriptions[_sbb][PlaquetteOrientation.LEFT]
        )
        # tlb, otb: Top-Left Basis, Other Basis
        tlb = _sbb if both_cubes_have_spatial_pipes_in_both_dimensions else _sbb.flipped()
        otb = tlb.flipped()
        left, right = (3, 2) if both_cubes_have_spatial_pipes_in_both_dimensions else (1, 4)
        return FrozenDefaultDict(
            {
                right: right_plaquette,
                left: left_plaquette,
                5: up_bulk_plaquettes[tlb][Orientation.VERTICAL],
                6: up_bulk_plaquettes[otb][Orientation.HORIZONTAL],
                7: down_bulk_plaquettes[otb][Orientation.HORIZONTAL],
                8: down_bulk_plaquettes[tlb][Orientation.VERTICAL],
            },
            default_value=RPNGDescription.empty(),
        )

    ############################################################
    #                         Hadamard                         #
    ############################################################

    ########################################
    #           Regular junction           #
    ########################################
    def get_temporal_hadamard_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.Template` instance needed to implement a
        transversal Hadamard gate applied on one logical qubit.
        """
        return QubitTemplate()

    def get_temporal_hadamard_rpng_descriptions(
        self, is_reversed: bool, z_orientation: Orientation = Orientation.HORIZONTAL
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a transversal Hadamard gate
        applied on one logical qubit.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_temporal_hadamard_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.Template` instance returned by this
            method.

        Arguments:
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            z_orientation: orientation of the ``Z`` observable at the beginning
                of the generated circuit description. The ``Z`` observable
                orientation will be flipped at the end of the returned circuit
                description, which is exactly the expected behaviour for a
                Hadamard transition.
                Used to compute the stabilizers that should be measured on the
                boundaries and in the bulk of the returned logical qubit
                description.

        Returns:
            a description of the plaquettes needed to implement a transversal
            Hadamard gate applied on one logical qubit.

        """
        # Generating plaquette descriptions we will need later.
        bulk_descriptions = self.get_bulk_hadamard_rpng_descriptions(is_reversed)
        two_body_descriptions = self.get_2_body_rpng_descriptions(is_reversed, hadamard=True)
        hbasis = Basis.Z if z_orientation == Orientation.HORIZONTAL else Basis.X
        vbasis = hbasis.flipped()
        return FrozenDefaultDict(
            {
                6: two_body_descriptions[vbasis][PlaquetteOrientation.UP],
                7: two_body_descriptions[hbasis][PlaquetteOrientation.LEFT],
                9: bulk_descriptions[vbasis][Orientation.HORIZONTAL],
                10: bulk_descriptions[hbasis][Orientation.VERTICAL],
                12: two_body_descriptions[hbasis][PlaquetteOrientation.RIGHT],
                13: two_body_descriptions[vbasis][PlaquetteOrientation.DOWN],
            },
            default_value=RPNGDescription.empty(),
        )

    def get_temporal_hadamard_plaquettes(
        self, is_reversed: bool, z_orientation: Orientation = Orientation.HORIZONTAL
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a transversal Hadamard gate applied on one
        logical qubit.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_temporal_hadamard_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.Template` instance returned by this
            method.

        Arguments:
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            z_orientation: orientation of the ``Z`` observable at the beginning
                of the generated circuit description. The ``Z`` observable
                orientation will be flipped at the end of the returned circuit
                description, which is exactly the expected behaviour for a
                Hadamard transition.
                Used to compute the stabilizers that should be measured on the
                boundaries and in the bulk of the returned logical qubit
                description.

        Returns:
            the plaquettes needed to implement a transversal Hadamard gate applied on one logical
            qubit.

        """
        return self._mapper(self.get_temporal_hadamard_rpng_descriptions)(
            is_reversed, z_orientation
        )

    ########################################
    #                X pipe                #
    ########################################
    def get_spatial_vertical_hadamard_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.RectangularTemplate` instance needed to
        implement a spatial Hadamard pipe between two logical qubits aligned on the ``X`` axis.
        """
        return QubitVerticalBorders()

    def get_spatial_vertical_hadamard_rpng_descriptions(
        self,
        top_left_basis: Basis,
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a Hadamard spatial transition
        between two neighbouring logical qubits aligned on the ``X`` axis.

        The Hadamard transition basically exchanges the ``X`` and ``Z`` logical
        observables between two neighbouring logical qubits aligned on the ``X``
        axis.

        Note:
            By convention, the hadamard-like transition is performed at the
            top-most plaquettes.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_spatial_vertical_hadamard_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            top_left_basis: basis of the top-left-most stabilizer.
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a Hadamard
            spatial transition between two neighbouring logical qubits aligned
            on the ``X`` axis.

        """
        # Generating plaquette descriptions we will need later.
        bulk_descriptions = self.get_bulk_rpng_descriptions(is_reversed, reset, measurement)
        two_body_descriptions = self.get_2_body_rpng_descriptions(is_reversed)
        bulk1, bulk2, bottom = self.get_spatial_x_hadamard_rpng_descriptions(
            top_left_basis, is_reversed, reset, measurement
        )
        # tlb: top-left basis, otb: other basis.
        tlb, otb = top_left_basis, top_left_basis.flipped()
        return FrozenDefaultDict(
            {
                2: two_body_descriptions[otb][PlaquetteOrientation.UP],
                3: bottom,
                5: bulk1,
                6: bulk2,
                7: bulk_descriptions[tlb][Orientation.VERTICAL],
                8: bulk_descriptions[otb][Orientation.HORIZONTAL],
            },
            default_value=RPNGDescription.empty(),
        )

    def get_spatial_vertical_hadamard_plaquettes(
        self,
        top_left_basis: Basis,
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a Hadamard spatial transition between two
        neighbouring logical qubits aligned on the ``X`` axis.

        The Hadamard transition basically exchanges the ``X`` and ``Z`` logical
        observables between two neighbouring logical qubits aligned on the ``X``
        axis.

        Note:
            By convention, the hadamard-like transition is performed at the
            top-most plaquettes.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_spatial_vertical_hadamard_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            top_left_basis: basis of the top-left-most stabilizer.
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            the plaquettes needed to implement a Hadamard spatial transition between two
            neighbouring logical qubits aligned on the ``X`` axis.

        """
        return self._mapper(self.get_spatial_vertical_hadamard_rpng_descriptions)(
            top_left_basis, is_reversed, reset, measurement
        )

    ########################################
    #                Y pipe                #
    ########################################
    def get_spatial_horizontal_hadamard_raw_template(self) -> RectangularTemplate:
        """Return the :class:`~tqec.templates.base.RectangularTemplate` instance needed to
        implement a spatial Hadamard pipe between two logical qubits aligned on the ``Y`` axis.
        """
        return QubitHorizontalBorders()

    def get_spatial_horizontal_hadamard_rpng_descriptions(
        self,
        top_left_basis: Basis,
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> FrozenDefaultDict[int, RPNGDescription]:
        """Return a description of the plaquettes needed to implement a Hadamard spatial transition
        between two neighbouring logical qubits aligned on the ``Y`` axis.

        The Hadamard transition basically exchanges the ``X`` and ``Z`` logical
        observables between two neighbouring logical qubits aligned on the ``Y``
        axis.

        Note:
            By convention, the hadamard-like transition is performed at the
            top-most plaquettes.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_spatial_horizontal_hadamard_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            top_left_basis: basis of the top-left-most stabilizer.
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            a description of the plaquettes needed to implement a Hadamard
            spatial transition between two neighbouring logical qubits aligned
            on the ``Y`` axis.

        """
        # Generating plaquette descriptions we will need later.
        bulk_descriptions = self.get_bulk_rpng_descriptions(is_reversed, reset, measurement)
        two_body_descriptions = self.get_2_body_rpng_descriptions(is_reversed)
        bulk1, bulk2, left = self.get_spatial_y_hadamard_rpng_descriptions(
            top_left_basis, is_reversed, reset, measurement
        )
        # tlb: top-left basis, otb: other basis.
        tlb, otb = top_left_basis, top_left_basis.flipped()
        return FrozenDefaultDict(
            {
                1: left,
                4: two_body_descriptions[otb][PlaquetteOrientation.RIGHT],
                5: bulk1,
                6: bulk2,
                7: bulk_descriptions[otb][Orientation.VERTICAL],
                8: bulk_descriptions[tlb][Orientation.HORIZONTAL],
            },
            default_value=RPNGDescription.empty(),
        )

    def get_spatial_horizontal_hadamard_plaquettes(
        self,
        top_left_basis: Basis,
        is_reversed: bool,
        reset: Basis | None = None,
        measurement: Basis | None = None,
    ) -> Plaquettes:
        """Return the plaquettes needed to implement a Hadamard spatial transition between two
        neighbouring logical qubits aligned on the ``Y`` axis.

        The Hadamard transition basically exchanges the ``X`` and ``Z`` logical
        observables between two neighbouring logical qubits aligned on the ``Y``
        axis.

        Note:
            By convention, the hadamard-like transition is performed at the
            top-most plaquettes.

        Warning:
            This method is tightly coupled with
            :meth:`FixedBoundaryConventionGenerator.get_spatial_horizontal_hadamard_raw_template`
            and the returned ``RPNG`` descriptions should only be considered
            valid when used in conjunction with the
            :class:`~tqec.templates.base.RectangularTemplate` instance returned
            by this method.

        Arguments:
            top_left_basis: basis of the top-left-most stabilizer.
            is_reversed: flag indicating if the plaquette schedule should be
                reversed or not. Useful to limit the loss of code distance when
                hook errors are not correctly oriented by alternating regular
                and reversed plaquettes.
            reset: basis of the reset operation performed on **internal**
                data-qubits. Defaults to ``None`` that translates to no reset
                being applied on data-qubits.
            measurement: basis of the measurement operation performed on
                **internal** data-qubits. Defaults to ``None`` that translates
                to no measurement being applied on data-qubits.

        Returns:
            the plaquettes needed to implement a Hadamard spatial transition between two
            neighbouring logical qubits aligned on the ``Y`` axis.

        """
        return self._mapper(self.get_spatial_horizontal_hadamard_rpng_descriptions)(
            top_left_basis, is_reversed, reset, measurement
        )



"""Defines the :class:`.CorrelationSurface` class and functions to find them in a ZX graph."""

from __future__ import annotations

from pyzx.graph.graph_s import GraphS
from pyzx.pauliweb import PauliWeb, multiply_paulis

from tqec.computation.correlation import CorrelationSurface, ZXEdge, ZXNode
from tqec.utils.enums import Basis


def correlation_surface_to_pauli_web(
    correlation_surface: CorrelationSurface, g: GraphS
) -> PauliWeb[int, tuple[int, int]]:
    """Convert the correlation surface to a Pauli web.

    Args:
        correlation_surface: The correlation surface to convert.
        g: The ZX graph the correlation surface is based on.

    Returns:
        A `PauliWeb` representation of the correlation surface.

    """
    half_edge_bases: dict[tuple[int, int], set[str]] = {}
    for edge in correlation_surface.span:
        u, v = edge.u, edge.v
        half_edge_bases.setdefault((u.id, v.id), set()).add(u.basis.value)
        half_edge_bases.setdefault((v.id, u.id), set()).add(v.basis.value)

    pauli_web = PauliWeb(g)
    for e, bases in half_edge_bases.items():
        pauli = "I"
        for basis in bases:
            pauli = multiply_paulis(pauli, basis)
        pauli_web.add_half_edge(e, pauli)
    return pauli_web


def pauli_web_to_correlation_surface(
    pauli_web: PauliWeb[int, tuple[int, int]],
) -> CorrelationSurface:
    """Create a correlation surface from a Pauli web."""
    span: set[ZXEdge] = set()
    half_edges: dict[tuple[int, int], str] = pauli_web.half_edges()
    while half_edges:
        (u, v), pauli_u = half_edges.popitem()
        pauli_v = half_edges.pop((v, u))
        if pauli_u == "Y":  # pragma: no cover
            span.add(ZXEdge.sorted(ZXNode(u, Basis.X), ZXNode(v, Basis.X)))
            span.add(ZXEdge.sorted(ZXNode(u, Basis.Z), ZXNode(v, Basis.Z)))
            continue
        span.add(ZXEdge.sorted(ZXNode(u, Basis(pauli_u)), ZXNode(v, Basis(pauli_v))))
    return CorrelationSurface(frozenset(span))



from collections.abc import Iterable
from fractions import Fraction

from pyzx import EdgeType, VertexType
from pyzx.graph.graph_s import GraphS

from tqec.interop.pyzx.positioned import PositionedZX
from tqec.utils.position import Position3D


def make_positioned_zx_graph(
    vertex_types: list[VertexType],
    positions: list[Position3D],
    edges: Iterable[tuple[int, int]] = (),
    hadamard_edges: Iterable[bool] = (),
    phases: dict[int, Fraction] = {},
    inputs: tuple[int, ...] = (),
    outputs: tuple[int, ...] = (),
) -> PositionedZX:
    g = GraphS()
    for i, vt in enumerate(vertex_types):
        g.add_vertex(vt)
        if i in phases:
            g.set_phase(i, phases[i])
    for (s, t), hadamard in zip(edges, hadamard_edges):
        g.add_edge((s, t), EdgeType.HADAMARD if hadamard else EdgeType.SIMPLE)
    g.set_inputs(inputs)
    g.set_outputs(outputs)
    return PositionedZX(g, {i: p for i, p in enumerate(positions)})



from abc import ABC, abstractmethod
from dataclasses import dataclass

from typing_extensions import override

from tqec.utils.exceptions import TQECError


class ScheduleFunction(ABC):
    """Interface for classes that can map schedules to other schedules."""

    @abstractmethod
    def __call__(self, input_schedule: int) -> int:
        """Transform the provided ``input_schedule`` to a new schedule and return it."""
        pass


@dataclass(frozen=True)
class ScheduleOffset(ScheduleFunction):
    """Maps a schedule by applying a relative offset to it."""

    offset: int

    @override
    def __call__(self, input_schedule: int) -> int:
        return input_schedule + self.offset


@dataclass(frozen=True)
class ScheduleConstant(ScheduleFunction):
    """Maps a schedule to a constant."""

    constant: int

    def __post_init__(self) -> None:
        if self.constant < 0:
            raise TQECError(
                "Cannot have a negative schedule. "
                f"The provided constant is negative: {self.constant}."
            )

    @override
    def __call__(self, input_schedule: int) -> int:
        return self.constant



"""Block graph that represents three logical CNOT gates compressed in spacetime."""

from tqec.computation.block_graph import BlockGraph
from tqec.computation.cube import ZXCube
from tqec.utils.enums import Basis
from tqec.utils.position import Position3D


def three_cnots(observable_basis: Basis | None = None) -> BlockGraph:
    """Create a block graph for three logical CNOT gates compressed in spacetime.

    The three CNOT gates are applied in the following order:

    .. code-block:: text

        q0: -@---@-
             |   |
        q1: -X-@-|-
               | |
        q2: ---X-X-

    Args:
        observable_basis: The observable basis that the block graph can support. If
            None, the block graph will have open ports. Otherwise, the ports will be
            filled with the given observable basis.

    Returns:
        A :py:class:`~tqec.computation.block_graph.BlockGraph` instance representing the
        three logical CNOT gates compressed in spacetime.

    """
    g = BlockGraph("Three CNOTs")
    nodes = [
        (Position3D(-1, 0, 0), "P", "Out_a"),
        (Position3D(0, 0, 0), "XXZ", ""),
        (Position3D(0, -1, 0), "P", "In_a"),
        (Position3D(0, 1, 0), "XXZ", ""),
        (Position3D(1, 0, 0), "ZXZ", ""),
        (Position3D(1, 1, 0), "ZXZ", ""),
        (Position3D(1, 0, -1), "P", "In_b"),
        (Position3D(1, 1, -1), "P", "In_c"),
        (Position3D(1, 0, 1), "ZXX", ""),
        (Position3D(1, 1, 1), "ZXX", ""),
        (Position3D(2, 1, 0), "P", "Out_c"),
        (Position3D(1, 0, 2), "P", "Out_b"),
    ]

    for pos, kind, label in nodes:
        g.add_cube(pos, kind, label)

    pipes = [
        (1, 0),
        (1, 2),
        (1, 3),
        (1, 4),
        (4, 6),
        (4, 8),
        (5, 3),
        (5, 9),
        (5, 10),
        (5, 7),
        (8, 11),
        (8, 9),
    ]
    for p0, p1 in pipes:
        g.add_pipe(nodes[p0][0], nodes[p1][0])

    if observable_basis == Basis.Z:
        g.fill_ports(
            {
                "In_a": ZXCube.from_str("XZZ"),
                "In_b": ZXCube.from_str("ZXZ"),
                "In_c": ZXCube.from_str("ZXZ"),
                "Out_a": ZXCube.from_str("ZXZ"),
                "Out_b": ZXCube.from_str("ZXZ"),
                "Out_c": ZXCube.from_str("ZXZ"),
            }
        )
    elif observable_basis == Basis.X:
        g.fill_ports(
            {
                "In_a": ZXCube.from_str("XXZ"),
                "In_b": ZXCube.from_str("ZXX"),
                "In_c": ZXCube.from_str("ZXX"),
                "Out_a": ZXCube.from_str("XXZ"),
                "Out_b": ZXCube.from_str("ZXX"),
                "Out_c": ZXCube.from_str("XXZ"),
            }
        )
    return g



"""Defines core classes and functions to represent and manipulate quantum circuits.

This package defines the core class :class:`~schedule.circuit.ScheduledCircuit`
that is used to represent a quantum circuit in the `tqec` library. It also
defines a few core functions:

- :func:`~.generate_circuit` that takes a :class:`~tqec.templates.base.Template`
  instance and a description of plaquettes via a
  :class:`~tqec.plaquette.plaquette.Plaquettes` instance and generates a
  :class:`~schedule.circuit.ScheduledCircuit` instance that corresponds to the
  circuit described.
- :func:`~.merge_scheduled_circuits` that is a function that helps merging
  several :class:`~schedule.circuit.ScheduledCircuit` instances containing gates
  that are potentially scheduled at the same time (but not on the same qubits).

Functions from this package are really the backbone of the :mod:`tqec`
library and are re-used in higher-level packages (such as :mod:`tqec.compile`).

"""



"""Defines the :py:class:`~tqec.computation.pipe.Pipe` class."""

from __future__ import annotations

from collections.abc import Generator
from dataclasses import dataclass
from typing import Any

from tqec.computation.cube import Cube, ZXCube
from tqec.utils.enums import Basis
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Direction3D, Position3D


@dataclass(frozen=True)
class PipeKind:
    """The kind of a pipe in the block graph.

    Looking at the head of the pipe, i.e. the side which has smaller position,
    the kind is determined by the wall bases there together with whether the
    pipe has a hadamard transition. The basis along the direction of the pipe
    should be None.

    Attributes:
        x: At the head of the pipe, looking at the pipe along the x-axis, the basis
           of the walls observed. Can be None if the pipe connects two cubes along
           the x-axis.
        y: At the head of the pipe, looking at the pipe along the y-axis, the basis
           of the walls observed. Can be None if the pipe connects two cubes along
           the y-axis.
        z: At the head of the pipe, looking at the pipe along the z-axis, the basis
           of the walls observed. Can be None if the pipe connects two cubes along
           the z-axis.
        has_hadamard: Whether the pipe has a hadamard transition.

    """

    x: Basis | None
    y: Basis | None
    z: Basis | None
    has_hadamard: bool = False

    def __post_init__(self) -> None:
        if sum(basis is None for basis in (self.x, self.y, self.z)) != 1:
            raise TQECError("Exactly one basis must be None for a pipe.")
        if len({self.x, self.y, self.z}) != 3:
            raise TQECError("Pipe must have different basis walls.")

    def __str__(self) -> str:
        return "".join(
            basis.value if basis is not None else "O" for basis in (self.x, self.y, self.z)
        ) + ("H" if self.has_hadamard else "")

    @staticmethod
    def from_str(string: str) -> PipeKind:
        """Create a pipe kind from the string representation.

        The string should be a 3-character or 4-character string. The first 3
        characters represent the basis of the walls at the head of the pipe
        along the x, y, and z axes. If there is no wall along an axis, i.e.
        the pipe connects two cubes along that axis, the character should be
        "O" for open boundary. The last character, if exists, should be "H"
        to indicate the pipe has a hadamard transition.

        Args:
            string: The string representation of the pipe kind.

        Returns:
            The pipe kind represented by the string.

        """
        string = string.upper()
        has_hadamard = len(string) == 4 and string[3] == "H"
        bases = list(Basis(s) if s != "O" else None for s in string[:3])
        return PipeKind(
            bases[0],
            bases[1],
            bases[2],
            has_hadamard=has_hadamard,
        )

    @property
    def direction(self) -> Direction3D:
        """The direction along which the pipe connects the cubes."""
        return Direction3D(str(self).index("O"))

    def get_basis_along(self, direction: Direction3D, at_head: bool = True) -> Basis | None:
        """Get the wall basis of the pipe in the specified direction.

        Args:
            direction: The direction along which to get the basis.
            at_head: If True, get the basis at the head of the pipe, i.e. the side
                which has smaller position. Otherwise, get the basis at the tail of
                the pipe. This matters when the pipe has hadamard transition.

        Returns:
            None if the direction is the same as the pipe direction. Otherwise, the
            basis of the wall in the specified direction.

        """
        if direction == self.direction:
            return None
        head_basis = Basis(str(self)[direction.value])
        if not at_head and self.has_hadamard:
            return head_basis.flipped()
        return head_basis

    @property
    def is_temporal(self) -> bool:
        """Verify whether the pipe is temporal, i.e. connects two cubes along the Z axis."""
        return self.z is None

    @property
    def is_spatial(self) -> bool:
        """Verify whether the pipe is spatial, i.e. connects two cubes along the X or Y axis."""
        return not self.is_temporal

    @staticmethod
    def _from_cube_kind(
        cube_kind: ZXCube,
        direction: Direction3D,
        cube_at_head: bool,
        has_hadamard: bool = False,
    ) -> PipeKind:
        """Infer the pipe kind from the endpoint cube kind and the pipe direction.

        Args:
            cube_kind: The kind of the cube at the endpoint of the pipe.
            direction: The direction of the pipe.
            cube_at_head: Whether the cube at the head of the pipe, i.e. the
                side with smaller position.
            has_hadamard: Whether the pipe has a hadamard transition.

        Returns:
            The inferred pipe kind.

        """
        bases: list[Basis]
        if not cube_at_head and has_hadamard:
            bases = [basis.flipped() for basis in cube_kind.as_tuple()]
        else:
            bases = list(cube_kind.as_tuple())
        bases_str = [basis.value for basis in bases]
        bases_str[direction.value] = "O"
        if has_hadamard:
            bases_str.append("H")
        return PipeKind.from_str("".join(bases_str))


@dataclass(frozen=True)
class Pipe:
    """A block connecting two :py:class:`~tqec.computation.cube.Cube` objects.

    The pipes connect the cubes in the block graph. Topologically, the pipes
    open the boundaries of the cubes and wire them together to form the topological
    structure representing the computation. In the circuit level, the pipes
    make modifications to the blocks of operations represented by the cubes.
    Importantly, the pipes themselves do not occupy spacetime volume.

    WARNING: The attributes `u` and `v` will be ordered to ensure the position of `u`
    is less than `v`.

    Attributes:
        u: The cube at the head of the pipe. The position of u will be guaranteed to be less than v.
        v: The cube at the tail of the pipe. The position of v will be guaranteed to be greater
            than u.
        kind: The kind of the pipe.

    """

    u: Cube
    v: Cube
    kind: PipeKind

    def __post_init__(self) -> None:
        u, v = self.u, self.v
        p1, p2 = u.position, v.position
        shift = [0, 0, 0]
        shift[self.kind.direction.value] = 1
        if not p1.shift_by(*shift) == p2 and not p2.shift_by(*shift) == p1:
            raise TQECError(
                f"The pipe must connect two nearby cubes in direction {self.kind.direction}."
            )

        # Ensure position of u is less than v
        if p1 > p2:
            object.__setattr__(self, "u", v)
            object.__setattr__(self, "v", u)

    @staticmethod
    def from_cubes(u: Cube, v: Cube) -> Pipe:
        """Construct a pipe connecting two cubes and infer the pipe kind.

        Args:
            u: The first cube.
            v: The second cube.

        Returns:
            The pipe connecting the two cubes.

        Raises:
            TQECError: If the cubes are not neighbours or both of them are not ZX cubes.
            TQECError: If kind of the pipe cannot be inferred from the cubes.

        """
        if not u.is_zx_cube and not v.is_zx_cube:
            raise TQECError("At least one cube must be a ZX cube to infer the pipe kind.")
        u, v = (u, v) if u.position < v.position else (v, u)
        if not u.position.is_neighbour(v.position):
            raise TQECError("The cubes must be neighbours to create a pipe.")
        direction = next(
            d
            for d in Direction3D.all_directions()
            if u.position.shift_in_direction(d, 1) == v.position
        )

        # One cube is not a ZX cube
        if not u.is_zx_cube or not v.is_zx_cube:
            infer_from = u if u.is_zx_cube else v
            cube_kind = infer_from.kind
            assert isinstance(cube_kind, ZXCube)
            pipe_kind = PipeKind._from_cube_kind(cube_kind, direction, infer_from == u)
            return Pipe(u, v, pipe_kind)

        # Both cubes are ZX cubes
        assert isinstance(u.kind, ZXCube) and isinstance(v.kind, ZXCube)
        has_hadamard = {
            u.kind.get_basis_along(d) != v.kind.get_basis_along(d)
            for d in Direction3D.all_directions()
            if d != direction
        }
        if len(has_hadamard) == 2:
            raise TQECError("Cannot infer a valid pipe kind from the cubes.")
        pipe_kind = PipeKind._from_cube_kind(u.kind, direction, True, has_hadamard.pop())
        return Pipe(u, v, pipe_kind)

    @property
    def direction(self) -> Direction3D:
        """The direction along which the pipe connects the cubes."""
        return self.kind.direction

    def at_head(self, position: Position3D) -> bool:
        """Whether the position is at the head (u) of the pipe."""
        if position == self.u.position:
            return True
        if position == self.v.position:
            return False
        raise TQECError(f"The position {position} is not an endpoint of the pipe.")

    def __iter__(self) -> Generator[Cube]:
        yield self.u
        yield self.v

    def __str__(self) -> str:
        return f"{self.u}--({self.kind})--{self.v}"

    def to_dict(self) -> dict[str, Any]:
        """Return the dictionary representation of the pipe."""
        return {
            "u": self.u.position.as_tuple(),
            "v": self.v.position.as_tuple(),
            "kind": str(self.kind),
        }
